# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import logging
import queue
import time
from threading import Event, Thread


_LOG = logging.getLogger(__name__)


class BSThreadDiedException(Exception):
    pass


class BSThread:
    def __init__(self, updater):
        self._queue = queue.Queue()
        self._exit_event = Event()

        args = (
            self._queue,
            self._exit_event,
            updater,
        )

        self._thread = Thread(target=_bs_thread_program, args=args, daemon=True)

    def start(self):
        self._thread.start()

    def put_data(self, data):
        if self._exit_event.is_set() is True or self._thread.is_alive() is False:
            self.close()
            raise BSThreadDiedException

        try:
            self._queue.put(data)
        except BrokenPipeError:
            self.close()
            raise BSThreadDiedException

    def close(self):
        if self._thread.is_alive() is True:
            self._exit_event.set()
            self._thread.join(1)

        if self._thread.is_alive() is True:
            raise RuntimeError


def _bs_thread_program(q, exit_event, updater):
    try:
        from blinkstick import blinkstick
    except ImportError:
        pass

    FIND_ATTEMPT_INTERVAL = 0.5
    UPDATE_INTERVAL = 1 / 30

    last_find_attempt_time = -1.0
    last_update_time = -1.0
    stick = None
    data = None
    data_index = -1

    while not exit_event.is_set():
        try:
            data = q.get(timeout=0.001)
        except queue.Empty:
            pass
        else:
            data_index += 1

        if stick is None:
            now = time.perf_counter()

            if (now - last_find_attempt_time) > FIND_ATTEMPT_INTERVAL:
                last_find_attempt_time = now

                try:
                    stick = blinkstick.find_first()
                    stick.turn_off()
                    stick.pulse(blue=100, duration=100, steps=10, repeats=2)
                except Exception:
                    stick = None

        if stick is None:
            continue

        now = time.perf_counter()

        if (now - last_update_time) < UPDATE_INTERVAL:
            continue

        last_update_time = now

        try:
            if data_index < 0:
                stick.turn_off()
            else:
                updater.update(data, data_index, now, stick)
        except blinkstick.BlinkStickException:
            stick = None

    if stick is not None:
        try:
            stick.turn_off()
        except Exception as e:
            _LOG.exception(e)

    try:
        while True:
            q.get(timeout=0.001)
    except Exception:  # noqa: S110
        pass


class _ExampleBSUpdater:
    def update(self, data, data_index, t, stick):
        x = float(data)
        stick.set_color(red=50 * (1 - x), index=0)
        stick.set_color(blue=50 * x, index=1)


if __name__ == "__main__":
    bs_updater = _ExampleBSUpdater()
    bs_thread = BSThread(bs_updater)
    bs_thread.start()

    t0 = time.monotonic()
    t = 0
    while t < 5:
        t = time.monotonic() - t0
        y = t % 1.0 > 0.5

        try:
            bs_thread.put_data(y)
        except BSThreadDiedException:
            exit()

        time.sleep(0.01)

    bs_thread.close()



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

from typing import Optional, Tuple

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import ENVELOPE_FWHM_M, PeakSortingMethod
from acconeer.exptool.a121.algo.distance._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
)


@attrs.frozen(kw_only=True)
class ProcessorSpec:
    processor_config: ProcessorConfig = attrs.field()
    group_index: int = attrs.field()
    subsweep_indexes: list[int] = attrs.field()
    processor_context: Optional[ProcessorContext] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class AggregatorConfig:
    peak_sorting_method: PeakSortingMethod = attrs.field(default=PeakSortingMethod.STRONGEST)


@attrs.frozen(kw_only=True)
class AggregatorResult:
    processor_results: list[ProcessorResult] = attrs.field()
    estimated_distances: npt.NDArray[np.float64] = attrs.field()
    estimated_strengths: npt.NDArray[np.float64] = attrs.field()
    near_edge_status: Optional[bool] = attrs.field(default=None)
    service_extended_result: list[dict[int, a121.Result]] = attrs.field()


class Aggregator:
    """Aggregating class

    Instantiates Processor objects based configuration from Detector.

    Aggregates result, based on selected peak sorting strategy, from underlying Processor objects.
    """

    def __init__(
        self,
        session_config: a121.SessionConfig,
        extended_metadata: list[dict[int, a121.Metadata]],
        config: AggregatorConfig,
        specs: list[ProcessorSpec],
        sensor_id: int,
    ):
        self.config = config
        self.specs = specs
        self.sensor_id = sensor_id

        self.processors: list[Processor] = []

        for spec in specs:
            metadata = extended_metadata[spec.group_index][self.sensor_id]
            sensor_config = session_config.groups[spec.group_index][self.sensor_id]

            processor = Processor(
                sensor_config=sensor_config,
                metadata=metadata,
                processor_config=spec.processor_config,
                subsweep_indexes=spec.subsweep_indexes,
                context=spec.processor_context,
            )
            self.processors.append(processor)

    def process(self, extended_result: list[dict[int, a121.Result]]) -> AggregatorResult:
        processors_result = []
        dists: npt.NDArray[np.float64] = np.array([])
        strengths: npt.NDArray[np.float64] = np.array([])
        profile_fwhms: npt.NDArray[np.float64] = np.array([])

        for spec, processor in zip(self.specs, self.processors):
            processor_result = processor.process(extended_result[spec.group_index][self.sensor_id])
            processors_result.append(processor_result)
            if processor_result.estimated_distances is not None:
                strengths = np.concatenate(
                    (strengths, np.array(processor_result.estimated_strengths))
                )
                dists = np.concatenate((dists, np.array(processor_result.estimated_distances)))
                profile_fwhms = np.concatenate(
                    (
                        profile_fwhms,
                        np.ones_like(np.array(processor_result.estimated_distances))
                        * ENVELOPE_FWHM_M[processor.profile],
                    )
                )
        # Remove first element as the FWHM of the subsweep with greater profile is used when
        # determining the distance for merging two peaks.
        if profile_fwhms.shape[0] != 0:
            profile_fwhms = np.delete(profile_fwhms, 0)

        (dists_merged, strengths_merged) = self._merge_peaks(profile_fwhms, dists, strengths)
        (dists_sorted, strengths_sorted) = self._sort_peaks(
            dists_merged, strengths_merged, self.config.peak_sorting_method
        )

        return AggregatorResult(
            processor_results=processors_result,
            estimated_distances=dists_sorted,
            estimated_strengths=strengths_sorted,
            near_edge_status=processors_result[0].near_edge_status,
            service_extended_result=extended_result,
        )

    @staticmethod
    def _merge_peaks(
        profile_fwhms: npt.NDArray[np.float64],
        dists: npt.NDArray[np.float64],
        strengths: npt.NDArray[np.float64],
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        sorting_order = np.argsort(dists)
        distances_sorted = dists[sorting_order]
        strengths_sorted = strengths[sorting_order]
        profile_fwhms = profile_fwhms[sorting_order[1:] - 1]

        peak_cluster_idxs = np.where(profile_fwhms < np.diff(distances_sorted))[0] + 1
        distances_merged = [
            np.mean(cluster)
            for cluster in np.split(distances_sorted, peak_cluster_idxs)
            if dists.size != 0
        ]
        strengths_merged = [
            np.mean(cluster)
            for cluster in np.split(strengths_sorted, peak_cluster_idxs)
            if dists.size != 0
        ]
        return (np.array(distances_merged), np.array(strengths_merged))

    @staticmethod
    def _sort_peaks(
        dists: npt.NDArray[np.float64],
        strengths: npt.NDArray[np.float64],
        method: PeakSortingMethod,
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        if method == PeakSortingMethod.CLOSEST:
            quantity_to_sort = dists
        elif method == PeakSortingMethod.STRONGEST:
            quantity_to_sort = -strengths
        else:
            msg = "Unknown peak sorting method"
            raise ValueError(msg)
        return (
            np.array([dists[i] for i in quantity_to_sort.argsort()]),
            np.array([strengths[i] for i in quantity_to_sort.argsort()]),
        )



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from .buffered_link import BufferedLink, LinkError
from .null_link import NullLink, NullLinkError
from .serial_link import ExploreSerialLink, SerialLink, SerialProcessLink
from .socket_link import SocketLink
from .usb_link import PyUsbCdc, USBLink



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from acconeer.exptool import a121

from ._example_app import ExampleAppConfig


def _get_base_config() -> ExampleAppConfig:
    return ExampleAppConfig(
        measured_point=80,
        profile=a121.Profile.PROFILE_3,
        hwaas=16,
        frame_rate=None,
        inter_frame_idle_state=a121.IdleState.READY,
        inter_sweep_idle_state=a121.IdleState.READY,
    )


def get_low_frequency_config() -> ExampleAppConfig:
    # Example app config for running low frequency(low sweep rate) in continuous sweep mode
    example_app_config = _get_base_config()
    example_app_config.continuous_sweep_mode = True
    example_app_config.double_buffering = True
    example_app_config.sweeps_per_frame = 20
    example_app_config.sweep_rate = 200
    example_app_config.low_frequency_enhancement = True
    example_app_config.lp_coeff = 0.8

    return example_app_config


def get_high_frequency_config() -> ExampleAppConfig:
    # Example app config for running frame mode
    example_app_config = _get_base_config()
    example_app_config.continuous_sweep_mode = False
    example_app_config.double_buffering = False
    example_app_config.sweeps_per_frame = 1024
    example_app_config.sweep_rate = 10000
    example_app_config.low_frequency_enhancement = False
    example_app_config.lp_coeff = 0.5

    return example_app_config



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import typing as t
from pathlib import Path

from PySide6.QtWidgets import (
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QPushButton,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from .syntax_highlight import SyntaxHighlighter


class _JsonEditor(QWidget):
    def __init__(self, submit_slot: t.Callable[[str], t.Any]) -> None:
        super().__init__()

        self.text_edit = QTextEdit()
        self.text_edit.setPlaceholderText("Input JSON here")
        self.text_edit.setFontFamily("monospace")
        self._syntax_highlight = SyntaxHighlighter.json(self.text_edit.document())

        self.load_button = QPushButton("Load JSON")
        self.load_button.clicked.connect(lambda: submit_slot(self.text_edit.toPlainText()))

        layout = QVBoxLayout()
        layout.addWidget(self.text_edit, stretch=1)
        layout.addWidget(self.load_button, stretch=0)
        self.setLayout(layout)


class LoadDialogWithJsonEditor(QDialog):
    def __init__(
        self,
        dialog: QFileDialog,
        parent: t.Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._contents: t.Optional[str] = None

        self._dialog = dialog
        self._dialog.accepted.connect(self._set_selected_file_contents)
        self._dialog.rejected.connect(self.reject)

        layout = QHBoxLayout()
        layout.addWidget(self._dialog, stretch=1)
        layout.addWidget(
            _JsonEditor(self._set_contents),
            stretch=1,
        )
        self.setLayout(layout)

    def _set_selected_file_contents(self) -> None:
        """Handles the "Open" button being pressed in the dialog"""
        (selected_file,) = self._dialog.selectedFiles()
        self._contents = Path(selected_file).read_text()
        self.accept()

    def _set_contents(self, contents: str) -> None:
        """Handles the submitted content from _JsonEditor."""
        self._contents = contents
        self.accept()

    @classmethod
    def get_load_contents(
        cls,
        caption: str = "",
        filter: str = "All (*)",
        options: QFileDialog.Option = QFileDialog.Option.DontUseNativeDialog,
        parent: t.Optional[QWidget] = None,
    ) -> t.Optional[str]:
        dialog = QFileDialog(caption=caption)
        dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        dialog.setNameFilter(filter)
        dialog.setOptions(options)

        instance = cls(dialog, parent=parent)
        if instance.exec():
            return instance._contents
        else:
            return None



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Optional

from PySide6 import QtCore
from PySide6.QtWidgets import QGridLayout, QGroupBox, QLabel, QLineEdit, QWidget

from acconeer.exptool import a121


class RangeHelpView(QGroupBox):
    APPROX_BASE_STEP_LENGTH_M = 2.5e-3

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setTitle("Approx. selected range")

        layout = QGridLayout(self)
        self.setLayout(layout)

        self._start = RangeHelpViewValueWidget(self)
        layout.addWidget(QLabel("Start", self), 0, 0)
        layout.addWidget(self._start, 1, 0)

        self._end = RangeHelpViewValueWidget(self)
        layout.addWidget(QLabel("End", self), 0, 1)
        layout.addWidget(self._end, 1, 1)

        self._step = RangeHelpViewValueWidget(self)
        layout.addWidget(QLabel("Step", self), 0, 2)
        layout.addWidget(self._step, 1, 2)

        self.set_data(None)

    def set_data(self, subsweep_config: Optional[a121.SubsweepConfig]) -> None:
        if subsweep_config:
            end_point = (
                subsweep_config.start_point
                + (subsweep_config.num_points - 1) * subsweep_config.step_length
            )
            start_m = subsweep_config.start_point * self.APPROX_BASE_STEP_LENGTH_M
            end_m = end_point * self.APPROX_BASE_STEP_LENGTH_M
            step_m = subsweep_config.step_length * self.APPROX_BASE_STEP_LENGTH_M

            self._start.setText(f"{start_m:.3f} m")
            self._end.setText(f"{end_m:.3f} m")
            self._step.setText(f"{step_m * 1e3:.1f} mm")
        else:
            self._start.clear()
            self._step.clear()
            self._end.clear()


class RangeHelpViewValueWidget(QLineEdit):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.setEnabled(False)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Optional

from acconeer.exptool.a121._core.entities import (
    SensorCalibration,
    SessionConfig,
)
from acconeer.exptool.a121._core.utils import iterate_extended_structure


def get_calibrations_provided(
    session_config: SessionConfig,
    calibrations: Optional[dict[int, SensorCalibration]] = None,
) -> dict[int, bool]:
    calibrations_provided = {}
    for _, sensor_id, _ in iterate_extended_structure(session_config.groups):
        if calibrations:
            calibrations_provided[sensor_id] = sensor_id in calibrations
        else:
            calibrations_provided[sensor_id] = False

    return calibrations_provided



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

from PySide6 import QtCore
from PySide6.QtGui import QRegularExpressionValidator
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app.new._enums import ConnectionInterface, ConnectionState, PluginGeneration
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.qt_subclasses import AppModelAwareWidget
from acconeer.exptool.app.new.ui.device_comboboxes import SerialPortComboBox, USBDeviceComboBox
from acconeer.exptool.app.new.ui.icons import COG, LINK, UNLINK


class _ConnectAndDisconnectButton(QPushButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setFixedWidth(120)

        app_model.sig_notify.connect(self._on_app_model_update)

        self.clicked.connect(self._on_click)

    def _on_click(self) -> None:
        if self.app_model.connection_state == ConnectionState.DISCONNECTED:
            self.app_model.connect_client()
        elif self.app_model.connection_state == ConnectionState.CONNECTED:
            self.app_model.disconnect_client()
        else:
            raise RuntimeError

    def _on_app_model_update(self, app_model: AppModel) -> None:
        TEXTS = {
            ConnectionState.DISCONNECTED: "Connect",
            ConnectionState.CONNECTING: "Connecting...",
            ConnectionState.CONNECTED: "Disconnect",
            ConnectionState.DISCONNECTING: "Disconnecting...",
        }
        ENABLED_STATES = {ConnectionState.CONNECTED, ConnectionState.DISCONNECTED}
        ICONS = {
            ConnectionState.DISCONNECTED: LINK(),
            ConnectionState.CONNECTING: LINK(),
            ConnectionState.CONNECTED: UNLINK(),
            ConnectionState.DISCONNECTING: UNLINK(),
        }
        TOOLTIPS = {
            ConnectionState.DISCONNECTED: "Connect to device using specified interface",
            ConnectionState.CONNECTING: "Connecting...",
            ConnectionState.CONNECTED: "Disconnect the device",
            ConnectionState.DISCONNECTING: "Disconnecting...",
        }

        self.setText(TEXTS[app_model.connection_state])
        self.setEnabled(
            app_model.connection_state in ENABLED_STATES
            and app_model.plugin_state.is_steady
            and app_model.is_connect_ready()
        )
        self.setIcon(ICONS[app_model.connection_state])
        self.setToolTip(TOOLTIPS[app_model.connection_state])


class _SocketConnectionWidget(AppModelAwareWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(app_model, parent)
        self.app_model = app_model

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.ip_line_edit = QLineEdit(self)
        self.ip_line_edit.setPlaceholderText("<IP address>")
        self.ip_line_edit.editingFinished.connect(self._on_line_edit)
        self.ip_line_edit.returnPressed.connect(self._on_return_pressed)
        self.ip_line_edit.setMinimumWidth(125)
        layout.addWidget(self.ip_line_edit)

        self.setLayout(layout)

    def _on_line_edit(self) -> None:
        self.app_model.set_socket_connection_ip(self.ip_line_edit.text())

    def _on_return_pressed(self) -> None:
        if self.app_model.connection_state == ConnectionState.DISCONNECTED:
            self.app_model.connect_client()

    def on_app_model_update(self, app_model: AppModel) -> None:
        if not self.ip_line_edit.isModified():
            self.ip_line_edit.setText(str(app_model.socket_connection_ip))


class _SerialConnectionWidget(AppModelAwareWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(app_model, parent)
        self.app_model = app_model

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addWidget(SerialPortComboBox(app_model, self))

        self.setLayout(layout)


class _ConnectSettingsButton(QPushButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        self.setFlat(True)
        self.app_model = app_model

        self.setIcon(COG())
        self.setToolTip("Advanced settings")

        self.settings_dialog = _ConnectSettingsDialog(app_model, self)

        app_model.sig_notify.connect(self._on_app_model_update)
        self.clicked.connect(self._on_click)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.connection_state == ConnectionState.DISCONNECTED)

    def _on_click(self) -> None:
        self.settings_dialog.exec()


class _ConnectSettingsBaudrate(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model
        app_model.sig_notify.connect(self._on_app_model_update)

        self.baudrate_line_edit = QLineEdit(self)
        validator = QRegularExpressionValidator(self)
        # Empty string or integer
        validator.setRegularExpression(QtCore.QRegularExpression("(^[0-9]+$|^$)"))
        self.baudrate_line_edit.setValidator(validator)
        self.baudrate_line_edit.setPlaceholderText("auto")
        self.baudrate_line_edit.textEdited.connect(self._on_text_edit)
        self.baudrate_line_edit.editingFinished.connect(self._on_line_edit)
        self.baudrate_line_edit.setFixedWidth(125)

        layout = QHBoxLayout()
        layout.addWidget(QLabel("Baudrate:"))
        layout.addWidget(self.baudrate_line_edit)

        self.setLayout(layout)
        self.setToolTip("Override default baudrate of serial device")

    def _on_text_edit(self) -> None:
        if not self.baudrate_line_edit.text():
            self.baudrate_line_edit.setStyleSheet("* {font: italic}")
        else:
            self.baudrate_line_edit.setStyleSheet("* {}")

    def _on_line_edit(self) -> None:
        baudrate_text = self.baudrate_line_edit.text()
        baudrate = None if baudrate_text == "" else int(baudrate_text)
        self.app_model.set_overridden_baudrate(baudrate)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        if not self.baudrate_line_edit.isModified():
            if app_model.overridden_baudrate is None:
                self.baudrate_line_edit.setText("")
            else:
                self.baudrate_line_edit.setText(str(app_model.overridden_baudrate))
            self._on_text_edit()


class _ConnectSettingsDialog(QDialog):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.auto_connect_check_box = QCheckBox("Auto-connect", self)
        if self.app_model.autoconnect_forced:
            self.auto_connect_check_box.setToolTip(
                "This option is overriden by the '--autoconnect' option"
            )
            self.auto_connect_check_box.setDisabled(True)
        else:
            self.auto_connect_check_box.setToolTip("Enables auto-connect of device")

            self.auto_connect_check_box.stateChanged.connect(self._auto_connect_on_state_changed)
            app_model.sig_notify.connect(self._on_app_model_update)

        self.flow_control_check_box = QCheckBox("Flow control", self)
        self.flow_control_check_box.stateChanged.connect(self._flow_control_on_state_changed)

        layout = QVBoxLayout(self)
        layout.addWidget(_ConnectSettingsBaudrate(app_model, self))
        layout.addWidget(self.auto_connect_check_box)
        layout.addWidget(self.flow_control_check_box)

        self.setWindowTitle("Advanced settings")
        self.setMinimumWidth(300)
        self.setLayout(layout)

    def _auto_connect_on_state_changed(self) -> None:
        self.app_model.set_autoconnect_enabled(self.auto_connect_check_box.isChecked())

    def _flow_control_on_state_changed(self) -> None:
        self.app_model.set_flow_control_enabled(self.flow_control_check_box.isChecked())

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.auto_connect_check_box.setChecked(app_model.autoconnect_enabled)
        self.flow_control_check_box.setChecked(app_model.flow_control_enabled)


class _SimulatedConnectionWidget(AppModelAwareWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(app_model, parent)
        self.app_model = app_model

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)


class ClientConnectionWidget(AppModelAwareWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(app_model, parent)

        self.app_model = app_model

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.interface_dd = QComboBox(self)
        self.interface_dd.setMinimumWidth(110)

        layout.addWidget(self.interface_dd)

        self.interface_dd.addItem("Socket", userData=ConnectionInterface.SOCKET)
        self.interface_dd.addItem("Serial", userData=ConnectionInterface.SERIAL)
        self.interface_dd.addItem("USB", userData=ConnectionInterface.USB)
        self.interface_dd.addItem("Simulated", userData=ConnectionInterface.SIMULATED)

        self.interface_dd.currentIndexChanged.connect(self._on_interface_dd_change)

        self.stacked = QStackedWidget(self)
        self.stacked.setStyleSheet("QStackedWidget {background-color: transparent;}")
        self.stacked.addWidget(_SocketConnectionWidget(app_model, self.stacked))
        self.stacked.addWidget(_SerialConnectionWidget(app_model, self.stacked))
        self.stacked.addWidget(USBDeviceComboBox(app_model, self.stacked))
        self.stacked.addWidget(_SimulatedConnectionWidget(app_model, self.stacked))
        layout.addWidget(self.stacked)

        layout.addWidget(_ConnectSettingsButton(app_model, self))
        layout.addWidget(_ConnectAndDisconnectButton(app_model, self))

        layout.addStretch(1)
        self.setLayout(layout)

    def _on_interface_dd_change(self) -> None:
        self.app_model.set_connection_interface(self.interface_dd.currentData())

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.stacked.setEnabled(
            app_model.connection_state == ConnectionState.DISCONNECTED,
        )
        self.interface_dd.setEnabled(
            app_model.connection_state == ConnectionState.DISCONNECTED,
        )

        interface_index = self.interface_dd.findData(app_model.connection_interface)
        if interface_index == -1:
            raise RuntimeError

        self.interface_dd.setCurrentIndex(interface_index)
        self.stacked.setCurrentIndex(interface_index)


class GenerationSelection(QComboBox):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.addItem("A121", PluginGeneration.A121)
        self.setEnabled(False)

        self.currentIndexChanged.connect(
            lambda _: app_model.set_plugin_generation(self.currentData())
        )
        app_model.sig_notify.connect(self.on_app_model_update)
        self.set_current_data(app_model.plugin_generation)

    def set_current_data(self, generation: PluginGeneration) -> None:
        idx = self.findData(generation)

        assert idx >= 0, f"Generation {generation} could not be found in this combobox"

        with QtCore.QSignalBlocker(self):
            self.setCurrentIndex(idx)

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.set_current_data(app_model.plugin_generation)
        self.setEnabled(False)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import abc
import logging
import re
from pathlib import Path
from typing import Any, Callable, Generic, Optional, TypeVar

import h5py
from packaging.version import Version

from acconeer.exptool import a121
from acconeer.exptool.app.new import (
    ApplicationClient,
    BackendLogger,
    BackendPlugin,
    GeneralMessage,
    HandledException,
    Message,
    PluginGeneration,
    PluginState,
    PluginStateMessage,
    ViewPluginBase,
    get_temp_h5_path,
    is_task,
)


log = logging.getLogger(__name__)


T = TypeVar("T")


class A121BackendPluginBase(Generic[T], BackendPlugin[T]):
    _live_client: Optional[a121.Client]
    _replaying_client: Optional[a121.Client]
    _opened_record: Optional[a121.H5Record]
    _started: bool = False
    _recorder: Optional[a121.H5Recorder] = None

    def __init__(
        self,
        callback: Callable[[Message], None],
        generation: PluginGeneration,
        key: str,
        use_app_client: bool = True,
    ) -> None:
        super().__init__(callback, generation, key)
        self._live_client = None
        self._replaying_client = None
        self._opened_record = None
        self._logger = BackendLogger.getLogger(__name__)
        self._use_app_client = use_app_client

    @is_task
    def load_from_file(
        self,
        *,
        path: Path,
        config_override: Optional[Any] = None,
        context_override: Optional[Any] = None,
    ) -> None:
        if config_override is not None:
            self._logger.warning(f"Ignoring config override of type: {type(config_override)}")
        if context_override is not None:
            self._logger.warning(f"Ignoring context override of type: {type(context_override)}")

        try:
            self._opened_record = a121.H5Record(h5py.File(path, mode="r"))

            log_version_match = re.match(
                "^([0-9]+[.][0-9]+[.][0-9]+)[.]", self._opened_record.lib_version
            )
            if log_version_match is None:
                log_version = "0.0.0"  # assume old/any version
            else:
                log_version = log_version_match.groups()[0]

            # This is to be able to replay files recorded before v7.0.0
            # when there were only one session per file
            if self._opened_record.num_sessions <= 1 and Version(log_version) < Version("7.0.0"):
                replaying_client = a121._ReplayingClient(self._opened_record, cycled_session_idx=0)
            else:
                replaying_client = a121._ReplayingClient(self._opened_record)

            self.load_from_record_setup(record=self._opened_record)
        except Exception as exc:
            self._opened_record = None
            self._replaying_client = None

            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            msg = "Could not load from file"

            # Add more details from exception thrown
            for exc_arg in exc.args:
                if isinstance(exc_arg, str):
                    msg += f", {exc_arg}"

            raise HandledException(msg) from exc
        if self._use_app_client:
            self._replaying_client = ApplicationClient.wrap_a121(replaying_client, self.callback)
        else:
            self._replaying_client = replaying_client

        self.start_session(with_recorder=False)

        self.send_status_file_access_message(f"{path.name}", opened=True)
        self.broadcast()

    @abc.abstractmethod
    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        pass

    @is_task
    def load_from_cache(self) -> None:
        try:
            with self.h5_cache_file() as f:
                self._load_from_cache(f)
        except FileNotFoundError:
            pass
        except Exception:
            log.warning("Cache file loading failed, using defaults")
            self.restore_defaults()

        self._sync_sensor_ids()
        self.broadcast()

    @abc.abstractmethod
    def _load_from_cache(self, file: h5py.File) -> None:
        pass

    @is_task
    @abc.abstractmethod
    def restore_defaults(self) -> None:
        pass

    @abc.abstractmethod
    def get_next(self) -> None:
        pass

    def idle(self) -> bool:
        if self._started:
            if self.client is None:
                msg = "Client is not attached. Can not 'get_next'."
                raise RuntimeError(msg)

            try:
                self.get_next()
            except (a121._StopReplay, a121.ReplaySessionsExhaustedError):
                self.stop_session()
                return True
            except Exception as exc:
                try:
                    self.stop_session()
                except Exception as e:
                    self._logger.exception(e)

                msg = "Failed to get_next"
                raise HandledException(msg) from exc

            return True
        else:
            return False

    @property
    def client(self) -> Optional[a121.Client]:
        if self._replaying_client is not None:
            return self._replaying_client

        return self._live_client

    @abc.abstractmethod
    def _sync_sensor_ids(self) -> None:
        pass

    def attach_client(self, *, client: a121.Client) -> None:
        if self._use_app_client:
            self._live_client = ApplicationClient.wrap_a121(client, self.callback)
        else:
            self._live_client = client
        self._sync_sensor_ids()
        self.broadcast()

    def detach_client(self) -> None:
        self._live_client = None

    @abc.abstractmethod
    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        pass

    @abc.abstractmethod
    def end_session(self) -> None:
        pass

    @is_task
    def start_session(self, *, with_recorder: bool = True) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            msg = "Client is not attached. Can not 'start'."
            raise RuntimeError(msg)

        if not self.client.connected:
            msg = "Client is not connected. Can not 'start'."
            raise RuntimeError(msg)

        self.callback(GeneralMessage(name="saveable_file", data=None))
        if with_recorder:
            self._recorder = a121.H5Recorder(get_temp_h5_path())
        else:
            self._recorder = None

        self.callback(PluginStateMessage(state=PluginState.LOADED_STARTING))
        try:
            self._start_session(self._recorder)
        except Exception as exc:
            msg = "Could not start"
            # Add more details from exception thrown
            for exc_arg in exc.args:
                if isinstance(exc_arg, str):
                    msg += f", {exc_arg}"

            recorder = self.client.detach_recorder()
            if recorder is not None:
                recorder.close()
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException(msg) from exc

        self._started = True

        self.broadcast()

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

    @is_task
    def stop_session(self) -> None:
        if not self._started:
            log.debug("Not stopping. Session was not started")
            return

        self.callback(PluginStateMessage(state=PluginState.LOADED_STOPPING))
        try:
            self.end_session()
        except Exception as exc:
            msg = "Failure when stopping session"
            raise HandledException(msg) from exc
        finally:
            if self._recorder is not None:
                assert self._recorder.path is not None
                path = Path(self._recorder.path)
                self.callback(GeneralMessage(name="saveable_file", data=path))
                self._recorder = None

            self._started = False
            self.broadcast()
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        if self._opened_record:
            self.send_status_file_access_message(
                f"{self._opened_record.file.filename}", opened=False
            )
            self._opened_record.close()
            self._opened_record = None
            self._replaying_client = None
            self._sync_sensor_ids()

    def teardown(self) -> None:
        try:
            with self.h5_cache_file(write=True) as file:
                self.save_to_cache(file)
        except Exception:
            log.warning("Could not write to cache")

        self.detach_client()

    @abc.abstractmethod
    def save_to_cache(self, file: h5py.File) -> None:
        pass


class A121ViewPluginBase(ViewPluginBase):
    def _send_start_request(self) -> None:
        A121BackendPluginBase.start_session.rpc(
            self.app_model.put_task,
            with_recorder=self.app_model.recording_enabled,
        )

    def _send_stop_request(self) -> None:
        A121BackendPluginBase.stop_session.rpc(self.app_model.put_task)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

import abc


class BufferedLink(abc.ABC):
    DEFAULT_TIMEOUT: float = 2.0

    def __init__(self) -> None:
        self._timeout = self.DEFAULT_TIMEOUT

    @property
    def timeout(self) -> float:
        """Return link timout."""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        """Set return link timeout."""
        self._timeout = timeout
        self._update_timeout()

    @abc.abstractmethod
    def _update_timeout(self) -> None:
        """Propagates the newly set timeout (found in self._timeout)"""
        pass

    @abc.abstractmethod
    def connect(self) -> None:
        """Establishes a connection."""
        pass

    @abc.abstractmethod
    def recv(self, num_bytes: int) -> bytes:
        """Recieves `num_bytes` bytes."""
        pass

    @abc.abstractmethod
    def recv_until(self, byte_sequence: bytes) -> bytes:
        """Collects all bytes until `byte_sequence` is encountered,
        returning what was collected
        """
        pass

    @abc.abstractmethod
    def send(self, bytes_: bytes) -> None:
        """Sends all `bytes_` over the link."""
        pass

    @abc.abstractmethod
    def disconnect(self) -> None:
        """Tears down the connection."""
        pass


class LinkError(RuntimeError):
    pass



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from enum import Enum, unique
from typing import Optional, TypeVar

from acconeer.exptool._core.class_creation.enum import (
    find_by_first_element_in_value,
    find_by_lowercase_name,
    find_by_name,
    find_by_value,
)


T = TypeVar("T", bound=Enum)


@unique
class Profile(Enum):
    """Profile"""

    PROFILE_1 = 1
    PROFILE_2 = 2
    PROFILE_3 = 3
    PROFILE_4 = 4
    PROFILE_5 = 5

    @classmethod
    def _missing_(cls, value: object) -> Optional[Profile]:
        return find_by_value(cls, value) or find_by_name(cls, value)


@unique
class PRF(Enum):
    """Pulse Repetition Frequency (PRF)

    Pulse Repetition Frequency, PRF, is the frequency at which pulses are sent out from
    the radar system. The measurement time is approximately proportional to the PRF.
    The higher the PRF, the shorter the measurement time.

    This parameter sets the Maximum Measurable Distance, MMD, that can be achieved. MMD is the
    maximum value for the end point, i.e., the start point + (number of points * step length).
    For example, an MMD of 7.0 m means that the range cannot be set further out than 7.0 m.

    It also sets the Maximum Unambiguous Range, MUR, that can be achieved. MUR is the maximum
    distance at which an object can be located to guarantee that its reflection corresponds to
    the most recent transmitted pulse. Objects farther away than the MUR may fold into the
    measured range. For example, with a MUR of 11.5 m, an object at 13.5 m could become
    visible at 2 m.

    ================= ======== ====== ======
    PRF Setting            PRF    MMD    MUR
    ================= ======== ====== ======
    PRF_19_5_MHZ [*]_ 19.5 MHz  3.1 m  7.7 m
    PRF_15_6_MHZ      15.6 MHz  5.1 m  9.6 m
    PRF_13_0_MHZ      13.0 MHz  7.0 m 11.5 m
    PRF_8_7_MHZ        8.7 MHz 12.7 m 17.3 m
    PRF_6_5_MHZ        6.5 MHz 18.5 m 23.1 m
    PRF_5_2_MHZ        5.2 MHz 24.3 m 28.8 m
    ================= ======== ====== ======

    .. [*] 19.5MHz is only available for profile 1.
    """

    #               PRF       MMD  MUR
    PRF_19_5_MHz = (19500000, 3.1, 7.7)
    PRF_15_6_MHz = (15600000, 5.1, 9.6)
    PRF_13_0_MHz = (13000000, 7.0, 11.5)
    PRF_8_7_MHz = (8700000, 12.7, 17.3)
    PRF_6_5_MHz = (6500000, 18.5, 23.1)
    PRF_5_2_MHz = (5200000, 24.3, 28.8)

    @property
    def frequency(self) -> int:
        return int(self.value[0])

    @property
    def maximum_measurable_distance(self) -> float:
        return float(self.value[1])

    @property
    def mmd(self) -> float:
        return self.maximum_measurable_distance

    @property
    def maximum_unambiguous_range(self) -> float:
        return float(self.value[2])

    @property
    def mur(self) -> float:
        return self.maximum_unambiguous_range

    @classmethod
    def _missing_(cls, value: object) -> Optional[PRF]:
        return find_by_first_element_in_value(cls, value) or find_by_name(cls, value)


@unique
class IdleState(Enum):
    """Idle state"""

    DEEP_SLEEP = 0
    SLEEP = 1
    READY = 2

    @classmethod
    def _missing_(cls, value: object) -> Optional[IdleState]:
        if find_by_value(cls, value) is not None:
            return find_by_value(cls, value)

        if find_by_name(cls, value) is not None:
            return find_by_name(cls, value)

        if find_by_lowercase_name(cls, value) is not None:
            return find_by_lowercase_name(cls, value)

        return None

    def is_deeper_than(self, other: IdleState) -> bool:
        return self.value < other.value



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import copy
import logging
from functools import partial
from typing import Any, Iterator, Optional

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QTabWidget, QToolButton, QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool._core.entities.validation_result import Criticality
from acconeer.exptool.app.new.ui.components import pidgets
from acconeer.exptool.app.new.ui.components.data_editor import DataEditor
from acconeer.exptool.app.new.ui.components.group_box import GroupBox
from acconeer.exptool.app.new.ui.components.json_save_load_buttons import (
    JsonButtonOperations,
    create_json_save_load_buttons_from_type,
)
from acconeer.exptool.app.new.ui.components.types import PidgetFactoryMapping

from .subsweep_config_editor import SubsweepConfigEditor


log = logging.getLogger(__name__)


class SensorConfigEditor(DataEditor[Optional[a121.SensorConfig]]):
    sig_update = Signal(object)

    _sensor_config: Optional[a121.SensorConfig]

    _sensor_config_pidgets: dict[str, pidgets.Pidget]
    _erroneous_aspects: set[str]
    _read_only: bool
    _supports_multiple_subsweeps: bool

    SPACING = 15
    IDLE_STATE_LABEL_MAP = {
        a121.IdleState.READY: "Ready",
        a121.IdleState.SLEEP: "Sleep",
        a121.IdleState.DEEP_SLEEP: "Deep sleep",
    }
    SENSOR_CONFIG_FACTORIES: PidgetFactoryMapping = {
        "sweeps_per_frame": pidgets.IntPidgetFactory(
            name_label_text="Sweeps per frame:",
            name_label_tooltip=(
                "The number of sweeps that will be captured in each frame (measurement)."
            ),
            limits=(1, 4095),
        ),
        "sweep_rate": pidgets.OptionalFloatPidgetFactory(
            name_label_text="Sweep rate:",
            name_label_tooltip=(
                "The sweep rate for sweeps in a frame (measurement).\n"
                "If 'Limit' is unchecked, the sweep rate will be as fast as possible."
            ),
            limits=(1, 1e6),
            decimals=0,
            init_set_value=1000.0,
            placeholder_text="- Hz",
            suffix="Hz",
            checkbox_label_text="Limit",
        ),
        "frame_rate": pidgets.OptionalFloatPidgetFactory(
            name_label_text="Frame rate:",
            name_label_tooltip=(
                "Frame rate.\nIf 'Limit' is unchecked, the rate is not limited by the sensor "
                "but by the rate that the server acknowledge and reads out the frame."
            ),
            limits=(0.1, 1e4),
            decimals=1,
            init_set_value=10.0,
            placeholder_text="- Hz",
            suffix="Hz",
            checkbox_label_text="Limit",
        ),
        "inter_sweep_idle_state": pidgets.EnumPidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter sweep idle state:",
            name_label_tooltip=(
                "The inter sweep idle state is the state the sensor idles in "
                "between each sweep in a frame."
            ),
            label_mapping=IDLE_STATE_LABEL_MAP,
        ),
        "inter_frame_idle_state": pidgets.EnumPidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter frame idle state:",
            name_label_tooltip=(
                "The inter frame idle state is the state the sensor idles in between each frame."
            ),
            label_mapping=IDLE_STATE_LABEL_MAP,
        ),
        "continuous_sweep_mode": pidgets.CheckboxPidgetFactory(
            name_label_text="Continuous sweep mode",
            name_label_tooltip=(
                "<p>With CSM, the sensor timing is set up to generate a continuous "
                "stream of sweeps, even if more than one sweep per frame is used. "
                "The interval between the last sweep in one frame to the first "
                "sweep in the next frame becomes equal to the interval between "
                "sweeps within a frame (given by the sweep rate).</p>"
                ""
                "<p>It ensures that:</p>"
                ""
                "<p>'frame rate' = 'sweep rate' / 'sweeps per frame'</p>"
                ""
                "While the frame rate parameter can be set to approximately "
                "satisfy this condition, using CSM is more precise."
                ""
                "<p>If only one sweep per frame is used, CSM has no use since a "
                "continuous stream of sweeps is already given (if a fixed frame "
                "rate is used).</p>"
                ""
                "<p>The main use for CSM is to allow reading out data at a slower "
                "rate than the sweep rate, while maintaining that sweep rate "
                "continuously.</p>"
                ""
                "<p>Note that in most cases, double buffering must be enabled to "
                "allow high rates without delays.</p>"
            ),
        ),
        "double_buffering": pidgets.CheckboxPidgetFactory(
            name_label_text="Double buffering",
            name_label_tooltip=(
                "Double buffering will split the sensor buffer in two halves. "
                "One half is used to read out the frame while sampling is done "
                "in the other half."
            ),
        ),
    }

    def __init__(
        self,
        supports_multiple_subsweeps: bool = False,
        json_button_operations: JsonButtonOperations = JsonButtonOperations(0),
        parent: Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._sensor_config_pidgets = {}
        self._erroneous_aspects = set()

        self._sensor_config = None

        self._read_only = False
        self._supports_multiple_subsweeps = supports_multiple_subsweeps

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.sensor_group_box = GroupBox.vertical(
            "Sensor parameters",
            right_header=create_json_save_load_buttons_from_type(
                self,
                a121.SensorConfig,
                operations=json_button_operations,
            ),
            parent=self,
        )
        self.sensor_group_box.layout().setSpacing(self.SPACING)
        layout.addWidget(self.sensor_group_box)

        for aspect, factory in self.SENSOR_CONFIG_FACTORIES.items():
            pidget = factory.create(self.sensor_group_box)
            self.sensor_group_box.layout().addWidget(pidget)

            pidget.sig_update.connect(partial(self._update_sensor_config_aspect, aspect))

            self._sensor_config_pidgets[aspect] = pidget

        self.subsweep_group_box = GroupBox.vertical("Subsweep parameters", parent=self)
        self.subsweep_group_box.layout().setSpacing(self.SPACING)
        layout.addWidget(self.subsweep_group_box)

        self._tab_widget = QTabWidget(self)
        self._tab_widget.setStyleSheet("QTabWidget::pane { padding: 5px;}")
        self.subsweep_group_box.layout().addWidget(self._tab_widget)

        if supports_multiple_subsweeps:
            self._tab_widget.setTabsClosable(True)
            self._tab_widget.tabCloseRequested.connect(self._remove_subsweep_config)

            self._plus_button = QToolButton(self)
            self._plus_button.setText("+")
            layout.addWidget(self._plus_button)
            self._plus_button.clicked.connect(self._add_subsweep_config)
            self._tab_widget.setCornerWidget(self._plus_button)
            self._tab_widget.cornerWidget().setMinimumSize(self._plus_button.sizeHint())

        self.setLayout(layout)

    def set_read_only(self, read_only: bool) -> None:
        self._read_only = read_only

        for pidget in self._sensor_config_pidgets.values():
            pidget.setEnabled(not read_only)

        if self._supports_multiple_subsweeps:
            self._tab_widget.setTabsClosable(not read_only)
            self._plus_button.setEnabled(not read_only)

        for editor in self._subsweep_config_editors:
            editor.set_read_only(read_only)

    def set_data(self, sensor_config: Optional[a121.SensorConfig]) -> None:
        if self._sensor_config == sensor_config:
            return

        self._sensor_config = sensor_config

        self.setEnabled(sensor_config is not None)
        if self._sensor_config is None:
            log.debug("could not update ui as SensorConfig is None")
            return

        for aspect, pidget in self._sensor_config_pidgets.items():
            if aspect in self._erroneous_aspects:
                continue
            pidget.set_data(getattr(self._sensor_config, aspect))

        while self._tab_widget.count() > self._sensor_config.num_subsweeps:
            self._tab_widget.removeTab(0)

        while self._tab_widget.count() < self._sensor_config.num_subsweeps:
            new_tab_idx = self._tab_widget.count()
            self._tab_widget.addTab(self._create_subsweep_config_editor(new_tab_idx), "")

        for tab_idx, subsweep_editor in enumerate(self._subsweep_config_editors):
            self._tab_widget.setTabText(tab_idx, f"{tab_idx + 1}      ")
            subsweep_editor.sig_update.disconnect()
            subsweep_editor.sig_update.connect(
                partial(self._update_subsweep_config_at_index, tab_idx)
            )

        for subsweep, editor in zip(self._sensor_config.subsweeps, self._subsweep_config_editors):
            editor.set_data(subsweep)

    def get_data(self) -> Optional[a121.SensorConfig]:
        return copy.deepcopy(self._sensor_config)

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        for aspect, pidget in self._sensor_config_pidgets.items():
            if aspect not in self._erroneous_aspects:
                pidget.set_note_text("")

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        for subsweep_config_editor in self._subsweep_config_editors:
            unhandled_results = subsweep_config_editor.handle_validation_results(unhandled_results)

        return unhandled_results

    @property
    def is_ready(self) -> bool:
        return self._erroneous_aspects == set() and all(
            se.is_ready for se in self._subsweep_config_editors
        )

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if result.aspect is None or self._sensor_config is None:
            return False

        if result.aspect in self._erroneous_aspects:
            # If there is an erroneous aspect in the GUI, we do not want to overwrite that.
            # Once the erroneous aspect is handled with, the same validation result will
            # come through here anyway
            return True

        result_handled = False

        if result.source == self._sensor_config:
            for aspect, pidget in self._sensor_config_pidgets.items():
                if result.aspect == aspect:
                    pidget.set_note_text(result.message, result.criticality)
                    result_handled = True

        return result_handled

    def _create_subsweep_config_editor(self, index: int) -> SubsweepConfigEditor:
        e = SubsweepConfigEditor()
        e.sig_update.connect(lambda ssc: self._update_subsweep_config_at_index(index, ssc))
        e.set_read_only(self._read_only)
        return e

    @property
    def _subsweep_config_editors(self) -> Iterator[SubsweepConfigEditor]:
        for tab_idx in range(self._tab_widget.count()):
            tab_widget = self._tab_widget.widget(tab_idx)
            assert isinstance(tab_widget, SubsweepConfigEditor)
            yield tab_widget

    def _add_subsweep_config(self) -> None:
        if self._sensor_config is None or self._sensor_config.num_subsweeps >= 4:
            return

        config = copy.deepcopy(self._sensor_config)
        config._subsweeps.append(a121.SubsweepConfig())
        self.set_data(config)
        self.sig_update.emit(config)

    def _remove_subsweep_config(self, idx: int) -> None:
        if self._sensor_config is None or len(self._sensor_config.subsweeps) <= 1:
            return

        config = copy.deepcopy(self._sensor_config)
        config._subsweeps.pop(idx)
        self.set_data(config)
        self.sig_update.emit(config)

    def _update_subsweep_config_at_index(
        self, index: int, subsweep_config: a121.SubsweepConfig
    ) -> None:
        if self._sensor_config is None or index not in range(self._sensor_config.num_subsweeps):
            return

        config = copy.deepcopy(self._sensor_config)
        config._subsweeps[index] = subsweep_config
        self.set_data(config)
        self.sig_update.emit(config)

    def _update_sensor_config_aspect(self, aspect: str, value: Any) -> None:
        if self._sensor_config is None:
            msg = "SensorConfig is None"
            raise TypeError(msg)

        config = copy.deepcopy(self._sensor_config)

        try:
            setattr(config, aspect, value)
        except Exception as e:
            self._sensor_config_pidgets[aspect].set_note_text(e.args[0], Criticality.ERROR)
            self._erroneous_aspects.add(aspect)

            # this emit needs to be done to signal that "is_ready" has changed.
            self.sig_update.emit(self.get_data())
        else:
            self._erroneous_aspects.discard(aspect)

            self.set_data(config)
            self.sig_update.emit(config)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6.QtWidgets import QPushButton, QVBoxLayout

from acconeer.exptool.app.new.ui.components import GroupBox


def create_plugin_selection_widget(
    title: str, *rows: t.Tuple[str, t.Iterable[t.Callable[[], t.Any]]]
) -> GroupBox[QVBoxLayout]:
    widget = GroupBox.vertical(left_header=title)

    widget.setStyleSheet("QPushButton { text-align: left; font-weight: bold; }")
    for button_label, on_clicks in rows:
        button = QPushButton(button_label)
        for on_click in on_clicks:
            button.clicked.connect(on_click)
        widget.layout().addWidget(button)

    return widget



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import abc
import typing as t
import uuid

import attrs

from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool.app.new.ui.components.collapsible_widget import CollapsibleWidget

from .common import MaybeIterable, as_sequence
from .pidgets import Pidget


WidgetHook = t.Callable[[QWidget, t.Mapping[str, Pidget]], None]


def _hooks_converter(a: MaybeIterable[WidgetHook]) -> t.Sequence[WidgetHook]:
    return as_sequence(a)


@attrs.frozen(kw_only=True, slots=False)
class PidgetGroup(abc.ABC):
    """The base pidget group."""

    _instance_id: uuid.UUID = attrs.field(factory=uuid.uuid4, init=False)
    """Unique ID for each instance. Enables using otherwise equal instances as hash keys"""

    hooks: t.Sequence[WidgetHook] = attrs.field(factory=tuple, converter=_hooks_converter)
    """Sequence of hooks for this instance"""

    @abc.abstractmethod
    def get_container(self, pidgets: t.Iterable[Pidget]) -> QWidget:
        """Wraps given pidgets in a container that is specified in each subclass"""


@attrs.frozen(kw_only=True, slots=False)
class FlatPidgetGroup(PidgetGroup):
    """A pidget group that only groups widgets hierarchically, not visually"""

    def get_container(self, pidgets: t.Iterable[Pidget]) -> QWidget:
        return _in_a_vboxed_widget(pidgets)


@attrs.frozen(kw_only=True, slots=False)
class CollapsiblePidgetGroup(PidgetGroup):
    """A pidget group that is collapsible"""

    label: str
    collapsed: bool

    def get_container(self, pidgets: t.Iterable[Pidget]) -> QWidget:
        w = CollapsibleWidget(self.label, _in_a_vboxed_widget(pidgets))
        w.set_collapsed(self.collapsed)
        return w


def _in_a_vboxed_widget(widgets: t.Iterable[QWidget]) -> QWidget:
    layout = QVBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(11)

    for widget in widgets:
        layout.addWidget(widget)

    dummy = QWidget()
    dummy.setLayout(layout)
    return dummy



# Copyright (c) Acconeer AB, 2024
# All rights reserved
from __future__ import annotations

import functools
import logging
import typing as t

from PySide6.QtWidgets import QWidget

from acconeer.exptool.app.new.app_model.app_model import AppModel
from acconeer.exptool.app.new.ui.misc import ExceptionWidget


_LOG = logging.getLogger(__name__)


class _UiMethodDecoratingMeta(type):
    """Metaclasses are to classes what classes are to instances.

    - A class defines how the instance should be created
    - A metaclass defines how the class should be created

    A metaclass is used to implement ``abc.ABC``, which checks that subclasses implement
    all methods previously marked with ``@abc.abstractmethod``.

    This metaclass wraps all callable members (i.e. defined methods) in
    subclasses with the ExceptionWidget context manager, making any Exception originating
    from plugins' ViewPlugin or PlotPlugin show as a pop-up instead of in the terminal.
    """

    @staticmethod
    def __try_unload_plugin_and_reraise(f: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        @functools.wraps(f)
        def wrapper(instance: t.Any, *args: t.Any, **kwargs: t.Any) -> t.Any:
            try:
                if isinstance(f, staticmethod):
                    return f(*args, **kwargs)
                else:
                    return f(instance, *args, **kwargs)
            except Exception as e:
                if isinstance(instance, UiPluginBase):
                    instance.panic()
                else:
                    _LOG.error(
                        f"First argument of {f} was not an instance of UiPluginBase (was {type(instance)}), cannot unload plugin."
                    )
                raise e

        return wrapper

    @staticmethod
    def __popup_on_exception(f: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        @functools.wraps(f)
        def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            with ExceptionWidget.context():
                return f(*args, **kwargs)

        return wrapper

    def __new__(cls, name: str, bases: tuple[type, ...], dct: dict[str, t.Any]) -> t.Any:
        return super().__new__(
            cls,
            name,
            bases,
            {
                name: (
                    cls.__popup_on_exception(
                        cls.__try_unload_plugin_and_reraise(
                            member,
                        ),
                    )
                    if callable(member)
                    else member
                )
                for name, member in dct.items()
            },
        )


# Jumping through loops to make above metaclass work with PySide
# T.y.
# https://stackoverflow.com/a/13183077 &
# A. Voitier @ https://bugreports.qt.io/browse/PYSIDE-1767

_QWidgetMeta = type(QWidget)


class _ShibokenObjectTypeFence(_QWidgetMeta): ...  # type: ignore[misc, valid-type]


class _UiPluginMeta(_ShibokenObjectTypeFence, _UiMethodDecoratingMeta, _QWidgetMeta): ...  # type: ignore[misc, valid-type]


class UiPluginBase(QWidget, metaclass=_UiPluginMeta):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()
        self.__app_model = app_model

    def panic(self) -> None:
        """Signals the AppModel to unload the current plugin (i.e. self)"""
        _LOG.error("Plugin panicked.")
        self.__app_model._unload_current_plugin()



# Copyright (c) Acconeer AB, 2023
# All rights reserved


import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.history_length_s = processing_config.history_length_s
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        max_num_of_sectors = max(6, self.depths.size // 3)
        self.sector_size = max(1, -(-self.depths.size // max_num_of_sectors))
        self.num_sectors = -(-self.depths.size // self.sector_size)
        self.sector_offset = (self.num_sectors * self.sector_size - self.depths.size) // 2

        self.setup_is_done = False
        self.latest_fpos = 0
        self.latest_fneg = 0

    def setup(self, win):
        win.setWindowTitle("Acconeer presence detection example")

        self.limit_lines = []
        dashed_pen = pg.mkPen("k", width=2.5, style=QtCore.Qt.DashLine)

        # Data plot

        self.data_plot = win.addPlot(
            row=0,
            col=0,
            title="Frame (blue), fast (orange), and slow (green)",
        )
        self.data_plot.setMenuEnabled(False)
        self.data_plot.setMouseEnabled(x=False, y=False)
        self.data_plot.hideButtons()
        self.data_plot.showGrid(x=True, y=True)
        self.data_plot.setLabel("bottom", "Distance (m)")
        self.data_plot.setLabel("left", "Amplitude")
        self.data_plot.setYRange(0, 2**16)
        self.frame_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(0),
        )
        self.fast_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.slow_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(2),
        )
        self.data_plot.addItem(self.frame_scatter)
        self.data_plot.addItem(self.fast_scatter)
        self.data_plot.addItem(self.slow_scatter)
        self.frame_smooth_limits = et.utils.SmoothLimits(self.sensor_config.update_rate)

        # Fast motion plot

        self.fast_plot = win.addPlot(
            row=1,
            col=0,
            title="Fast motion score",
        )
        self.fast_plot.setMenuEnabled(False)
        self.fast_plot.setMouseEnabled(x=False, y=False)
        self.fast_plot.hideButtons()
        self.fast_plot.showGrid(x=True, y=True)
        self.fast_plot.setLabel("bottom", "Distance (m)")
        self.fast_plot.setLabel("left", "Norm. ampl.")
        self.fast_curve = self.fast_plot.plot()
        self.fast_smooth_max = et.utils.SmoothMax(self.sensor_config.update_rate)
        zero_curve = self.fast_plot.plot(self.depths, np.zeros_like(self.depths))
        self.fast_smooth_max = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )
        fbi = pg.FillBetweenItem(
            zero_curve,
            self.fast_curve,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.fast_plot.addItem(fbi)
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.fast_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        self.move_depth_line_fast = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line_fast.hide()
        self.fast_plot.addItem(self.move_depth_line_fast)

        # Slow motion plot

        self.slow_plot = win.addPlot(
            row=2,
            col=0,
            title="Slow motion score",
        )
        self.slow_plot.setMenuEnabled(False)
        self.slow_plot.setMouseEnabled(x=False, y=False)
        self.slow_plot.hideButtons()
        self.slow_plot.showGrid(x=True, y=True)
        self.slow_plot.setLabel("bottom", "Distance (m)")
        self.slow_plot.setLabel("left", "Norm. ampl.")
        cal_implementation_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Adapting Threshold . . .")
        )
        self.cal_implementation_item = pg.TextItem(
            html=cal_implementation_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )
        zero_curve = self.slow_plot.plot(self.depths, np.zeros_like(self.depths))
        self.slow_curve = self.slow_plot.plot()
        self.threshold_curve = self.slow_plot.plot(pen=dashed_pen)
        self.slow_smooth_max = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line_slow = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line_slow.hide()
        self.slow_plot.addItem(self.move_depth_line_slow)
        fbi = pg.FillBetweenItem(
            zero_curve,
            self.slow_curve,
            brush=et.utils.pg_brush_cycler(0),
        )
        self.slow_plot.addItem(fbi)

        self.scatter_oject_curve = pg.ScatterPlotItem(
            size=15,
            brush=et.utils.pg_brush_cycler(3),
        )
        self.slow_plot.addItem(self.scatter_oject_curve)
        self.slow_plot.addItem(self.cal_implementation_item)
        self.cal_implementation_item.hide()

        # Presence history plot

        self.move_hist_plot = pg.PlotItem(title="Detection history")
        self.move_hist_plot.setMenuEnabled(False)
        self.move_hist_plot.setMouseEnabled(x=False, y=False)
        self.move_hist_plot.hideButtons()
        self.move_hist_plot.showGrid(x=True, y=True)
        self.move_hist_plot.setLabel("bottom", "Time (s)")
        self.move_hist_plot.setLabel("left", "Detection")
        self.move_hist_plot.getAxis("left").setTicks([[(0, "False"), (1, "True")]])
        self.move_hist_plot.setXRange(-self.history_length_s, 0)
        self.history_smooth_max = et.utils.SmoothMax(self.sensor_config.update_rate)
        self.move_hist_plot.setYRange(-0.1, 2)
        self.fneg_text_item = pg.TextItem(color="black", anchor=(0.5, 0))
        self.fpos_text_item = pg.TextItem(color="black", anchor=(0.5, 0))
        self.move_hist_slow_curve = self.move_hist_plot.plot(pen=et.utils.pg_pen_cycler())
        self.move_hist_fast_curve = self.move_hist_plot.plot(pen=et.utils.pg_pen_cycler(1))
        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )
        self.move_hist_plot.addItem(self.fneg_text_item)
        self.move_hist_plot.addItem(self.fpos_text_item)
        self.move_hist_plot.addItem(self.present_text_item)
        self.move_hist_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        # Sector plot

        self.sector_plot = pg.PlotItem()
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

        self.sectors.reverse()

        sublayout = win.addLayout(row=3, col=0)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_hist_plot, col=0)
        sublayout.addItem(self.sector_plot, col=1)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.data_plot.setVisible(self.processing_config.show_data)
        self.fast_plot.setVisible(self.processing_config.show_fast)
        self.slow_plot.setVisible(self.processing_config.show_slow)
        self.sector_plot.setVisible(self.processing_config.show_sectors)

        for line in self.limit_lines:
            line.setPos(processing_config.fast_motion_threshold)

    def update(self, data):
        self.frame_scatter.setData(
            np.tile(self.depths, self.sensor_config.sweeps_per_frame),
            data["frame"].flatten(),
        )

        self.fast_scatter.setData(self.depths, data["fast"])
        self.slow_scatter.setData(self.depths, data["slow"])
        self.data_plot.setYRange(*self.frame_smooth_limits.update(data["frame"]))

        fast = data["fast_motion"]
        self.fast_curve.setData(self.depths, fast)
        m_fast = self.fast_smooth_max.update(np.max(fast))
        m_fast = max(m_fast, 2 * self.processing_config.fast_motion_threshold)
        self.fast_plot.setYRange(0, m_fast)

        slow = data["slow_motion"]
        self.slow_curve.setData(self.depths, slow)
        m_slow = self.slow_smooth_max.update(np.max(slow))
        m_slow = max(m_slow, 2 * self.processing_config.slow_motion_threshold)
        self.slow_plot.setYRange(0, m_slow)
        threshold_array = data["threshold_array"]
        index_threshold_array = np.nonzero(
            [
                threshold_array
                > np.percentile([np.min(threshold_array), np.max(threshold_array)], 30)
            ]
        )
        self.depths_scatter_object = [self.depths[itx] for itx in list(index_threshold_array[1])]
        self.scatter_object = np.ones(len(self.depths_scatter_object)) * 0.5
        self.scatter_oject_curve.setData(self.depths_scatter_object, self.scatter_object)
        index_dps = np.argmax(slow)
        index_array_based = int(index_dps) == int(data["presence_distance_index"])
        self.threshold_curve.setData(self.depths, data["threshold_array"])
        movement_x = data["presence_distance"]
        self.move_depth_line_slow.setPos(movement_x)
        self.move_depth_line_fast.setPos(movement_x)
        self.move_depth_line_slow.setVisible(
            bool(data["presence_detected"] and data["information"] == "slow_motion")
        )
        self.move_depth_line_fast.setVisible(
            bool(data["presence_detected"] and data["information"] == "fast_motion")
        )

        if data["information"] == "adapting_threshold":
            self.cal_implementation_item.show()
            self.cal_implementation_item.setPos(1, 0.95 * m_slow)
        else:
            self.cal_implementation_item.hide()

        move_hist_slow_ys = data["slow_history"]
        move_hist_fast_ys = data["fast_history"]
        move_hist_xs = np.linspace(-self.history_length_s, 0, len(move_hist_slow_ys))
        self.move_hist_slow_curve.setData(move_hist_xs, move_hist_slow_ys)
        self.move_hist_fast_curve.setData(move_hist_xs, move_hist_fast_ys)
        self.set_present_text_y_pos(2)
        if self.latest_fneg == data["fls_neg_s"] and self.latest_fneg > 0:
            fneg_text = "False negative is > {:.2f} s".format(self.latest_fneg)
        else:
            fneg_text = "False negative is {:.2f} s".format(self.latest_fneg)
        if self.latest_fpos == data["fls_pos_s"] and self.latest_fpos > 0:
            fpos_text = "False positive is > {:.2f} s".format(self.latest_fpos)
        else:
            fpos_text = "False positive is {:.2f} s".format(self.latest_fpos)
        self.latest_fneg = data["fls_neg_s"]
        self.latest_fpos = data["fls_pos_s"]
        self.fpos_text_item.setText(fpos_text)
        self.fneg_text_item.setText(fneg_text)
        if data["presence_detected"] and index_array_based:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)
            self.present_text_item.show()
            self.not_present_text_item.hide()

        elif data["presence_detected"]:
            present_text = "Presence detected"
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        brush = et.utils.pg_brush_cycler(0)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data["presence_detected"]:
            index = (data["presence_distance_index"] + self.sector_offset) // self.sector_size
            self.sectors[index].setBrush(et.utils.pg_brush_cycler(1))

    def set_present_text_y_pos(self, y):
        self.present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)
        self.not_present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)
        self.fpos_text_item.setPos(-self.history_length_s * 0.9, 0.98 * y)
        self.fneg_text_item.setPos(-self.history_length_s * 0.9, 0.8 * y)



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from .erroneus_message import ErroneousMessage
from .log_message import LogMessage
from .message import Message, ParseError
from .set_baudrate_response import SetBaudrateResponse
from .streaming_responses import StartStreamingResponse, StopStreamingResponse
from .system_info_response import SystemInfoResponse



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import enum


class Mode(enum.Enum):
    POWER_BINS = "power_bins"
    ENVELOPE = "envelope"
    IQ = "iq"
    SPARSE = "sparse"


def get_mode(mode):
    if mode is None:
        return None

    if isinstance(mode, Mode):
        return mode

    if isinstance(mode, str):
        try:
            return Mode(mode.strip().lower())
        except ValueError:
            pass

        try:
            return Mode.__members__[mode.strip().upper()]
        except KeyError:
            pass

    msg = "unknown mode"
    raise ValueError(msg)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import enum
import itertools
import json
from functools import wraps
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import packaging.version

from acconeer.exptool._core import version_parsing


S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")

KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")

Number = Union[int, float]


def unextend(structure: list[dict[int, T]]) -> T:
    """'Unexpands' a structure and returns the single element in the structure."""
    try:
        (group,) = structure
        (entry,) = group.values()
        return entry
    except Exception as e:
        msg = f"Could not unextend the structure {structure}"
        raise ValueError(msg) from e


def convert_value(value: Number, *, factory: Callable[[Number], T]) -> T:
    try:
        # May raise ValueError if e.g. "value" is a non-int string
        converted_value = factory(value)

        if converted_value != value:
            # E.g. int("3") != "3", int(3.5) != 3.5. is catched here.
            raise ValueError

        return converted_value
    except ValueError:
        msg = f"{value} cannot be converted with {factory}"
        raise TypeError(msg)


def _check_bounds(
    value: Number,
    lower_bound: Optional[Number] = None,
    upper_bound: Optional[Number] = None,
    inclusive: bool = True,
) -> None:
    """Raises a ValueError if:
    * ``value`` is not in [lower_bound, upper_bound], if inclusive = True
    * ``value`` is not in (lower_bound, upper_bound), if inclusive = False
    """
    exclusive = not inclusive

    boundaries = f"{lower_bound}, {upper_bound}"
    interval = f"({boundaries})" if exclusive else f"[{boundaries}]"
    error = ValueError(f"{value} needs to be in {interval}")

    if lower_bound is not None and (
        (exclusive and value <= lower_bound) or (inclusive and value < lower_bound)
    ):
        raise error

    if upper_bound is not None and (
        (exclusive and value >= upper_bound) or (inclusive and value > upper_bound)
    ):
        raise error


def convert_validate_int(
    value: Union[float, int], max_value: Optional[int] = None, min_value: Optional[int] = None
) -> int:
    """Converts an argument to an int.

    :param value: The argument to be converted and boundary checked
    :param max_value: Maximum value allowed
    :param min_value: Minimum value allowed

    :raises: TypeError if value is a string or a float with decimals
    :raises: ValueError if value does not agree with max_value and min_value
    """
    int_value = convert_value(value, factory=int)
    _check_bounds(int_value, lower_bound=min_value, upper_bound=max_value, inclusive=True)
    return int_value


def validate_float(
    value: float,
    max_value: Optional[float] = None,
    min_value: Optional[float] = None,
    inclusive: bool = True,
) -> float:
    """Converts an argument to a float.

    :param value: The argument to be converted and boundary checked
    :param max_value: Maximum value allowed
    :param min_value: Minimum value allowed
    :param inclusive:
        Whether the bounds ``max_value`` and ``min_value`` should be considered inclusive.
        E.g. value = 0.0, min_value = 0.0, inclusive = False raises a ValueError.

    :raises: TypeError if value cannot be converted to a float.
    :raises: ValueError if value does not agree with max_value and min_value
    """
    float_value = convert_value(value, factory=float)
    _check_bounds(float_value, lower_bound=min_value, upper_bound=max_value, inclusive=inclusive)
    return float_value


def is_multiple_of(multiplier: int, product: int) -> bool:
    """Returns True if `product` is a multiple of `multiplier`.
    I.e. checks if `multiplicand` is an integer in the equation

    `multiplicand` * `multiplier` = `product`
    """
    return product >= multiplier and product % multiplier == 0


def is_divisor_of(divisor: int, dividend: int) -> bool:
    """Returns True if `dividend` is divided by `divisor` with no remainder
    I.e. checks that `quotient` is an integer in the equation

    `dividend` / `divisor` = `quotient`
    """
    return is_multiple_of(dividend, divisor)


def map_over_extended_structure(
    func: Callable[[ValueT], T], structure: list[dict[KeyT, ValueT]]
) -> list[dict[KeyT, T]]:
    """Applies a function, `func`, to each element of the extended structure.

    Example:

    >>> structure = [{1: "one"}, {2: "two"}]            # KeyT = int, ValueT = str
    >>> func = str.encode                               # ValueT = str, T = bytes
    >>> map_over_extended_structure(func, structure)
    [{1: b'one'}, {2: b'two'}]
    """
    return [{k: func(v) for k, v in d.items()} for d in structure]


def zip_extended_structures(
    structure_a: list[dict[int, S]],
    structure_b: list[dict[int, T]],
) -> list[dict[int, Tuple[S, T]]]:
    """Zip structures according to group id and sensor id.

    Typical use case is reordering extended structures before iterating.

    Example:

    >>> a = [{1: "a_1", 2: "a_2"}, {1: "a_3"}]
    >>> b = [{1: "b_1", 2: "b_2"}, {1: "b_3"}]
    >>> zip_extended_structures(a, b)
    [{1: ('a_1', 'b_1'), 2: ('a_2', 'b_2')}, {1: ('a_3', 'b_3')}]
    """
    res = []
    try:
        for group_id, sensor_id, value in iterate_extended_structure(structure_a):
            res.append(
                (
                    group_id,
                    sensor_id,
                    (value, structure_b[group_id][sensor_id]),
                ),
            )
    except (KeyError, IndexError):
        msg = "Structure of arguments are not the same."
        raise ValueError(msg)

    return create_extended_structure(iter(res))


def zip3_extended_structures(
    structure_a: list[dict[int, S]],
    structure_b: list[dict[int, T]],
    structure_c: list[dict[int, U]],
) -> list[dict[int, Tuple[S, T, U]]]:
    """Zip structures according to group id and sensor id.

    Typical use case is reordering extended structures before iterating.

    Example:

    >>> a = [{1: "a_1", 2: "a_2"}, {1: "a_3"}]
    >>> b = [{1: "b_1", 2: "b_2"}, {1: "b_3"}]
    >>> c = [{1: "c_1", 2: "c_2"}, {1: "c_3"}]
    >>> zip3_extended_structures(a, b, c)
    [{1: ('a_1', 'b_1', 'c_1'), 2: ('a_2', 'b_2', 'c_2')}, {1: ('a_3', 'b_3', 'c_3')}]
    """
    res = []
    try:
        for group_id, sensor_id, value in iterate_extended_structure(structure_a):
            res.append(
                (
                    group_id,
                    sensor_id,
                    (value, structure_b[group_id][sensor_id], structure_c[group_id][sensor_id]),
                ),
            )
    except (KeyError, IndexError):
        msg = "Structure of arguments are not the same."
        raise ValueError(msg)

    return create_extended_structure(iter(res))


def iterate_extended_structure(
    structure: Sequence[Mapping[int, ValueT]],
) -> Iterator[Tuple[int, int, ValueT]]:
    """Iterates over the elements of the extended structure.

    :returns: Iterator of (<group id>, <sensor id>, <element>)
    """

    for group_id, group in enumerate(structure):
        for sensor_id, elem in group.items():
            yield (group_id, sensor_id, elem)


def iterate_extended_structure_as_entry_list(
    structure: list[dict[int, ValueT]],
) -> Iterator[Tuple[int, int, ValueT]]:
    """Iterates over the elements of the extended structure.

    Note: this differs from ``iterate_extended_structure`` as it returns the
    ``entry_idx`` instead of the sensor id of a given entry.

    :returns: Iterator of (<group idx>, <entry idx>, <element>)
    """

    for group_idx, group in enumerate(structure):
        for entry_idx, elem in enumerate(group.values()):
            yield (group_idx, entry_idx, elem)


def extended_structure_entry_count(structure: list[dict[int, Any]]) -> int:
    """Traverses the extended structure and returns a count of all its entries

    :returns: number of entries
    """

    return sum(len(group) for group in structure)


def iterate_extended_structure_values(
    structure: Sequence[Mapping[int, ValueT]],
) -> Iterator[ValueT]:
    """Iterates like `iterate_extended_structure` but throws away group id and sensor id."""
    for _, _, value in iterate_extended_structure(structure):
        yield value


def extended_structure_shape(structure: list[dict[int, Any]]) -> list[set[int]]:
    return [set(group.keys()) for group in structure]


def transpose_extended_structures(
    structures: list[list[dict[int, ValueT]]],
) -> list[dict[int, list[ValueT]]]:
    """'Transposes' a list of extended structures to create an extended structure of lists"""

    if not structures:
        msg = "'structures' cannot be empty"
        raise ValueError(msg)

    shapes = [extended_structure_shape(s) for s in structures]
    if not all(shape == shapes[0] for shape in shapes):
        msg = "All extended structures needs to have the same structure."
        raise ValueError(msg)

    product: list[dict[int, list[ValueT]]] = map_over_extended_structure(
        lambda _: list(), structures[0]
    )

    for group_idx, sensor_id, value in itertools.chain(
        *[iterate_extended_structure(es) for es in structures]
    ):
        product[group_idx][sensor_id].append(value)

    return product


def create_extended_structure(items: Iterable[Tuple[int, int, ValueT]]) -> list[dict[int, ValueT]]:
    structure: list[dict[int, ValueT]] = []
    current_group_index: Optional[int] = None
    current_group: Optional[dict[int, ValueT]] = None

    for group_index, sensor_id, value in items:
        if current_group_index is None:
            if group_index != 0:
                raise ValueError

            current_group_index = 0
            current_group = {}
            structure.append(current_group)
        elif group_index != current_group_index:
            if group_index != current_group_index + 1:
                raise ValueError

            current_group_index += 1
            current_group = {}
            structure.append(current_group)

        assert current_group is not None
        if sensor_id in current_group:
            raise ValueError

        current_group[sensor_id] = value

    return structure


class EntityJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, enum.Enum):
            return obj.name

        return json.JSONEncoder.default(self, obj)


def parse_rss_version(rss_version: str) -> packaging.version.Version:
    """Takes an RSS version string and returns a corresponding Version

    The RSS version string is on a 'git describe'-like format:

        a121-vA.B.C<-rcD><-E-gF>
    """
    (release_line, version) = version_parsing.parse_rss_version(rss_version)

    if release_line != "a121":
        msg = "Not a valid RSS A121 version"
        raise ValueError(msg)

    return version


def indent_strs(strs: list[str], level: int) -> list[str]:
    return ["  " * level + s for s in strs]


def no_dynamic_member_creation(cls: Type[T]) -> Type[T]:
    """
    Class annotation that prevents setting any attributes that does not exist
    after the object have been created (the __init__ function have executed).
    """

    def setattr_wrapper(func):  # type: ignore
        @wraps(func)
        def setattr(self, key, value):  # type: ignore
            if hasattr(self, "__frozen") and not hasattr(self, key):
                msg = f'Invalid attribute "{key}" for {self!r}'
                raise AttributeError(msg)
            else:
                func(self, key, value)

        return setattr

    def init_wrapper(func):  # type: ignore
        @wraps(func)
        def init(self, *args, **kwargs):  # type: ignore
            func(self, *args, **kwargs)
            self.__frozen = True

        return init

    cls.__setattr__ = setattr_wrapper(cls.__setattr__)  # type: ignore
    cls.__init__ = init_wrapper(cls.__init__)  # type: ignore
    return cls



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QPushButton

from acconeer.exptool.app.new.ui.icons import GAUGE


_TOOL_TIP = """
Copies the current config into a new input block in the resource tab.

Note that no changes in the resource tab are not copied back here.
""".strip()


class GotoResourceTabButton(QPushButton):
    def __init__(self) -> None:
        super().__init__("View in Resource Calculator")
        self.setIcon(GAUGE())
        self.setIconSize(QSize(24, 24))
        self.setToolTip(_TOOL_TIP)



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import typing as t


_T = t.TypeVar("_T")
MaybeIterable = t.Union[_T, t.Iterable[_T]]


def as_sequence(a: MaybeIterable[_T]) -> t.Sequence[_T]:
    """
    Converts a single element or a iterable to a sequence

    Examples:

    >>> as_sequence('a')
    ('a',)

    >>> as_sequence(['a', 'b'])
    ('a', 'b')
    """
    if isinstance(a, t.Iterable):
        return tuple(a)
    else:
        return (a,)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import datetime
import os
import typing as t
from pathlib import Path
from uuid import uuid4

import h5py

import acconeer.exptool
from acconeer.exptool._core.entities import ClientInfo
from acconeer.exptool._core.recording import recorder
from acconeer.exptool._core.recording.h5_session_schema import SessionSchema
from acconeer.exptool.utils import get_module_version

from .saver import H5Saver
from .utils import PathOrH5File, h5_file_factory


_ConfigT = t.TypeVar("_ConfigT", contravariant=True)
_MetadataT = t.TypeVar("_MetadataT", contravariant=True)
_ResultT = t.TypeVar("_ResultT", contravariant=True)
_ServerInfoT = t.TypeVar("_ServerInfoT", contravariant=True)


def get_h5py_str_dtype() -> t.Any:
    return h5py.special_dtype(vlen=str)


_H5PY_STR_DTYPE = get_h5py_str_dtype()


def get_timestamp() -> str:
    return datetime.datetime.now().isoformat(timespec="seconds")


def get_uuid() -> str:
    return str(uuid4())


class H5Recorder(recorder.Recorder[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]):
    """Recorder writing directly to an HDF5 file

    :param path_or_file:
        If a path-like object, by default (depending on ``mode`` below) an HDF5 file is created at
        that path. When stopping, the file is closed.

        If an ``h5py.File``, that file is used as-is. In this case, the file is not closed when
        stopping.
    :param generation:
        Sensor generation being recorded
    :param saver:
        Saver used to write results to file
    :param attachable:
        A ``Client`` that the recorder will be attached to. When a recorder is attached to a
        ``Client``, all metadata and data frames will be recorded.
    :param mode:
        The file mode to use if a path-like object was given for ``path_or_file``. Default value is
        'x', meaning that we open for exclusive creation, failing if the file already exists.
    """

    _schema = SessionSchema

    path: t.Optional[os.PathLike[t.Any]]
    """The file path, if a path-like object was given for ``path_or_file``."""
    file: h5py.File
    """The ``h5py.File``."""
    owns_file: bool
    """Whether :class:`H5Recorder` opened and owns the file, i.e., if a path-like object was given
    for ``path_or_file``. If it does, the file is closed when stopping.
    """
    _saver: H5Saver[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]
    """Defining how session data should be stored in H5 session groups"""

    def __init__(
        self,
        path_or_file: PathOrH5File,
        generation: str,
        saver: H5Saver[_ConfigT, _MetadataT, _ResultT, _ServerInfoT],
        attachable: t.Optional[
            recorder.RecorderAttachable[
                recorder.Recorder[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]
            ]
        ] = None,
        mode: str = "x",
        *,
        _lib_version: t.Optional[str] = None,
        _timestamp: t.Optional[str] = None,
        _uuid: t.Optional[str] = None,
    ) -> None:
        self.file, self.owns_file = h5_file_factory(path_or_file, h5_file_mode=mode)
        self.path = Path(self.file.filename) if self.owns_file else None
        self._saver = saver
        self._current_session_group: t.Optional[h5py.Group] = None

        if attachable is None:
            self._attachable = None
        else:
            self._attachable = attachable
            attachable.attach_recorder(self)

        if _lib_version is None:
            _lib_version = get_module_version(acconeer.exptool)

        if _timestamp is None:
            _timestamp = get_timestamp()

        if _uuid is None:
            _uuid = get_uuid()

        self.file.create_dataset(
            "uuid",
            data=_uuid,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "timestamp",
            data=_timestamp,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "lib_version",
            data=_lib_version,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "generation",
            data=generation,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

    def _start(
        self,
        *,
        client_info: ClientInfo,
        server_info: _ServerInfoT,
    ) -> None:
        if "client_info" in self.file or "server_info" in self.file:
            msg = "It's not allowed to call '_start' once."
            raise RuntimeError(msg)

        self.file.create_dataset(
            "client_info",
            data=client_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        self._saver._write_server_info(self.file, server_info)

        self._saver._start()

    def _start_session(self, *, config: _ConfigT, metadata: _MetadataT, **kwargs: t.Any) -> None:
        self._current_session_group = self._schema.create_next_session_group(self.file)

        self._saver._start_session(
            self._current_session_group, config=config, metadata=metadata, **kwargs
        )

    def _sample(self, result: _ResultT) -> None:
        if self._current_session_group is None:
            msg = "No session group selected yet. This should not happen."
            raise RuntimeError(msg)

        self._saver._sample(self._current_session_group, results=[result])

    def _stop_session(self) -> None:
        self._saver._stop_session(self._current_session_group)

    def close(self) -> t.Any:
        try:
            self._stop_session()

            # If the client has self as an attached recorder, make sure to detatch it.
            if self._attachable is not None:
                _ = self._attachable.detach_recorder()
        finally:
            if self.owns_file:
                self.file.close()

    def require_algo_group(self, key: str) -> h5py.Group:
        """Creates/gets the ``algo`` group with a given ``key``.

        :raises: Exception if the key does not match the file content
        """

        group = self.file.require_group("algo")

        if "key" in group:
            existing_key = bytes(group["key"][()]).decode()
            if existing_key != key:
                msg = f"Algo group key mismatch: got '{key}' but had '{existing_key}'"
                raise Exception(msg)
        else:
            group.create_dataset(
                "key",
                data=key,
                dtype=_H5PY_STR_DTYPE,
                track_times=False,
            )

        return group



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import acconeer.exptool as et
from acconeer.exptool.a111.algo.presence_detection_sparse import _processor as presence_processing


def get_sensor_config():
    sensor_config = et.a111.SparseServiceConfig()
    sensor_config.range_interval = [0.24, 1.20]
    sensor_config.update_rate = 60
    sensor_config.sampling_mode = et.a111.SparseServiceConfig.SamplingMode.A
    sensor_config.profile = et.a111.SparseServiceConfig.Profile.PROFILE_3
    sensor_config.hw_accelerated_average_samples = 60
    return sensor_config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 2

    history_length = et.configbase.IntParameter(
        label="History length",
        default_value=100,
    )

    show_data_history_plot = et.configbase.BoolParameter(
        label="Show data history",
        default_value=True,
        updateable=True,
        order=110,
    )

    show_move_history_plot = et.configbase.BoolParameter(
        label="Show movement history",
        default_value=True,
        updateable=True,
        order=120,
    )


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        num_sensors = len(sensor_config.sensor)
        num_depths = et.a111.get_range_depths(sensor_config, session_info).size
        history_len = processing_config.history_length

        pd_config = presence_processing.ProcessingConfiguration()
        processor_class = presence_processing.Processor

        try:
            self.pd_processors = []
            for _ in sensor_config.sensor:
                p = processor_class(sensor_config, pd_config, session_info)
                self.pd_processors.append(p)
        except AssertionError:
            self.pd_processors = None

        self.data_history = np.ones([history_len, num_sensors, num_depths]) * 2**15
        self.presence_history = np.zeros([history_len, num_sensors, num_depths])

    def process(self, data, data_info):
        if self.pd_processors:
            if data_info is None:
                processed_datas = [p.process(s, None) for s, p in zip(data, self.pd_processors)]
            else:
                processed_datas = [
                    p.process(s, i) for s, i, p in zip(data, data_info, self.pd_processors)
                ]

            presences = [d["depthwise_presence"] for d in processed_datas]

            self.presence_history = np.roll(self.presence_history, -1, axis=0)
            self.presence_history[-1] = presences

        self.data_history = np.roll(self.data_history, -1, axis=0)
        self.data_history[-1] = data.mean(axis=1)

        out_data = {
            "data": data,
            "data_history": self.data_history,
            "presence_history": self.presence_history,
        }

        return out_data



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import itertools
import logging
from enum import Enum, auto
from typing import Callable, Dict, List, Optional, Type, TypeVar

import numpy as np
import numpy.typing as npt

from PySide6 import QtCore
from PySide6.QtGui import QTransform
from PySide6.QtWidgets import (
    QVBoxLayout,
)

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121 import algo
from acconeer.exptool.a121._core import utils as core_utils
from acconeer.exptool.a121.algo._plugins import (
    ExtendedProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
    ProcessorPluginSpec,
    ProcessorViewPluginBase,
    SetupMessage,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PidgetFactoryMapping,
    PlotPluginBase,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    backend,
    pidgets,
)
from acconeer.exptool.app.new.ui.components import GotoResourceTabButton, TabPGWidget

from ._processor import (
    AmplitudeMethod,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)


log = logging.getLogger(__name__)


_T = TypeVar("_T")


class PluginPresetId(Enum):
    DEFAULT = auto()


SEMI_TRANSPARENT_BRUSH = pg.mkBrush(color=(0xFF, 0xFF, 0xFF, int(0.8 * 0xFF)))


class BackendPlugin(ExtendedProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):
    PLUGIN_PRESETS = {
        PluginPresetId.DEFAULT.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_sensor_config()),
            processor_config=BackendPlugin.get_processor_config_cls()(),
        ),
    }

    @classmethod
    def get_processor(cls, state: ProcessorBackendPluginSharedState[ProcessorConfig]) -> Processor:
        return Processor(
            session_config=state.session_config,
            processor_config=state.processor_config,
        )

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)

        goto_resource_button = GotoResourceTabButton()
        goto_resource_button.clicked.connect(
            lambda: app_model.sig_resource_tab_input_block_requested.emit(
                self.session_config_editor.get_data()
            )
        )
        self.scrolly_layout.insertWidget(0, goto_resource_button)

    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "amplitude_method": pidgets.EnumPidgetFactory(
                enum_type=AmplitudeMethod,
                name_label_text="Amplitude method:",
                name_label_tooltip=(
                    "The method used to calculate the amplitude from the complex Sparse IQ data"
                ),
                label_mapping={
                    AmplitudeMethod.COHERENT: "Coherent",
                    AmplitudeMethod.NONCOHERENT: "Non-coherent",
                    AmplitudeMethod.FFT_MAX: "FFT Max",
                },
            )
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def supports_multiple_subsweeps(self) -> bool:
        return True

    @classmethod
    def supports_multiple_sensors(self) -> bool:
        return True


_Extended = List[Dict[int, _T]]


class PlotPlugin(PlotPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)

        self.smooth_max = et.utils.SmoothMax()
        self._plot_job: Optional[ProcessorResult] = None
        self._is_setup = False
        self.ampl_plot: Optional[pg.PlotItem] = None

        self.ampl_curves: _Extended[list[pg.PlotDataItem]] = []
        self.subsweeps_distances_m: _Extended[list[npt.NDArray[np.float64]]] = []

        layout = QVBoxLayout()

        self.amplitude_plot_widget = pg.GraphicsLayoutWidget()

        self.tab_widget = TabPGWidget()

        layout.addWidget(self.amplitude_plot_widget, stretch=1)
        layout.addWidget(self.tab_widget, stretch=2)

        self.setLayout(layout)
        self.tab_widget.setVisible(False)

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            if isinstance(message.metadata, a121.Metadata):
                metadata = [{message.session_config.sensor_id: message.metadata}]
            else:
                metadata = message.metadata

            self.setup(
                metadatas=metadata,
                session_config=message.session_config,
            )
            self._is_setup = True
            self.tab_widget.setVisible(True)
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(processor_result=self._plot_job)
        finally:
            self._plot_job = None

    def draw_plot_job(self, processor_result: ProcessorResult) -> None:
        if self.ampl_plot is None:
            raise RuntimeError

        max_ = 0.0

        for (
            plot_data_items,
            entry_result,
            subsweeps_distances_m,
        ) in core_utils.iterate_extended_structure_values(
            core_utils.zip3_extended_structures(
                core_utils.zip3_extended_structures(
                    self.ampl_curves, self.phase_curves, self.ft_images
                ),
                processor_result,
                self.subsweeps_distances_m,
            )
        ):
            for (
                amplitude_curve,
                phase_curve,
                ft_image,
                subsweep_result,
                subsweep_distances_m,
            ) in zip(*plot_data_items, entry_result, subsweeps_distances_m):
                ampls = subsweep_result.amplitudes
                amplitude_curve.setData(subsweep_distances_m, ampls)
                max_ = max(max_, np.max(ampls).item())

                phase_curve.setData(subsweep_distances_m, subsweep_result.phases)

                dvm = subsweep_result.distance_velocity_map
                ft_image.updateImage(dvm.T, levels=(0, 1.05 * np.max(dvm)))

        self.ampl_plot.setYRange(0, self.smooth_max.update(max_))

    def setup(
        self, metadatas: list[dict[int, a121.Metadata]], session_config: a121.SessionConfig
    ) -> None:
        self.amplitude_plot_widget.ci.clear()
        self.tab_widget.clear()

        self.subsweeps_distances_m = core_utils.map_over_extended_structure(
            lambda args: self._get_distances_m(*args),
            core_utils.zip_extended_structures(session_config.groups, metadatas),
        )

        self._setup_amplitude(session_config)
        self._setup_phase_and_dvm(session_config, metadatas)

    @staticmethod
    def _get_distances_m(
        config: a121.SensorConfig, metadata: a121.Metadata
    ) -> list[npt.NDArray[np.float64]]:
        return [algo.get_distances_m(subsweep, metadata) for subsweep in config.subsweeps]

    def _setup_amplitude(self, session_config: a121.SessionConfig) -> None:
        counter = itertools.count(0)

        entry_serial_numbers = core_utils.map_over_extended_structure(
            lambda _: next(counter), session_config.groups
        )
        shape = core_utils.extended_structure_shape(session_config.groups)
        sensor_ids = [{sensor_id: sensor_id for sensor_id in group} for group in shape]
        group_idxs = [
            {sensor_id: group_idx for sensor_id in group} for group_idx, group in enumerate(shape)
        ]

        num_unique_sensors = len(set(core_utils.iterate_extended_structure_values(sensor_ids)))

        config_placement = core_utils.zip3_extended_structures(
            entry_serial_numbers, sensor_ids, group_idxs
        )

        self.ampl_curves = core_utils.map_over_extended_structure(
            lambda args: self._create_amplitude_curves_for_subsweeps(
                args[0],
                *args[1],
                num_groups=len(session_config.groups),
                num_sensors=num_unique_sensors,
            ),
            core_utils.zip_extended_structures(session_config.groups, config_placement),
        )

        self.ampl_plot = self.amplitude_plot_widget.addPlot()
        self.ampl_plot.setMenuEnabled(False)
        self.ampl_plot.showGrid(x=True, y=True)
        self.ampl_plot.setLabel("left", "Amplitude")
        self.ampl_plot.setLabel("bottom", "Distance (m)")
        if (
            len(session_config.groups) > 1
            or num_unique_sensors > 1
            or any(
                sensor_config.num_subsweeps > 1
                for sensor_config in core_utils.iterate_extended_structure_values(
                    session_config.groups
                )
            )
        ):
            self.ampl_plot.addLegend(brush=SEMI_TRANSPARENT_BRUSH)

        for curves in core_utils.iterate_extended_structure_values(self.ampl_curves):
            for curve in curves:
                self.ampl_plot.addItem(curve)

    @staticmethod
    def _legend_label(
        sensor_id: int,
        group_idx: int,
        subsweep_index: int,
        num_groups: int,
        num_sensors: int,
        num_subsweeps: int,
    ) -> str:
        parts = []

        if num_groups > 1:
            if num_sensors > 1 or num_subsweeps > 1:
                parts += [f"G{group_idx}"]
            else:
                parts += [f"Group {group_idx}"]

        if num_sensors > 1:
            if num_groups > 1 or num_subsweeps > 1:
                parts += [f"S{sensor_id}"]
            else:
                parts += [f"Sensor {sensor_id}"]

        if num_subsweeps > 1:
            if num_groups > 1 or num_sensors > 1:
                parts += [f"SS{subsweep_index + 1}"]
            else:
                parts += [f"Subsweep {subsweep_index + 1}"]

        return ":".join(parts)

    def _create_amplitude_curves_for_subsweeps(
        self,
        sensor_config: a121.SensorConfig,
        entry_serial_number: int,
        sensor_id: int,
        group_idx: int,
        *,
        num_groups: int,
        num_sensors: int,
    ) -> list[pg.PlotDataItem]:
        subsweep_styles = [
            QtCore.Qt.PenStyle.SolidLine,
            QtCore.Qt.PenStyle.DashLine,
            QtCore.Qt.PenStyle.DotLine,
            QtCore.Qt.PenStyle.DashDotLine,
        ]

        return [
            pg.PlotDataItem(
                name=self._legend_label(
                    sensor_id,
                    group_idx,
                    index,
                    num_groups,
                    num_sensors,
                    sensor_config.num_subsweeps,
                ),
                pen=et.utils.pg_pen_cycler(entry_serial_number, style=subsweep_styles[index]),
            )
            for index, subsweep in enumerate(sensor_config.subsweeps)
        ]

    def _setup_phase_and_dvm(
        self, session_config: a121.SessionConfig, metadatas: list[dict[int, a121.Metadata]]
    ) -> None:
        phase_curves = []
        dvm_images = []

        for group_idx, sensor_id, sensor_config in core_utils.iterate_extended_structure(
            session_config.groups
        ):
            plot_widget = self.tab_widget.newPlotWidget(f"G{group_idx}:S{sensor_id}")

            phase_plot = plot_widget.addPlot(colspan=sensor_config.num_subsweeps)
            phase_plot.setMenuEnabled(False)
            phase_plot.showGrid(x=True, y=True)
            phase_plot.setLabel("left", "Phase")
            phase_plot.setYRange(-np.pi, np.pi)
            phase_plot.getAxis("left").setTicks(et.utils.pg_phase_ticks)

            if sensor_config.num_subsweeps > 1:
                phase_plot.addLegend(brush=SEMI_TRANSPARENT_BRUSH)

            subsweep_styles = [
                dict(symbol="o", symbolSize=5),  # circle
                dict(symbol="t1", symbolSize=5),  # triangle
                dict(symbol="s", symbolSize=4),  # square
                dict(symbol="d", symbolSize=6),  # diamond
            ]

            subsweep_phase_curves = [
                pg.PlotDataItem(
                    name=f"Subsweep {i + 1}",
                    pen=None,
                    **subsweep_styles[i],
                    symbolBrush=et.utils.pg_brush_cycler(0),
                    symbolPen="k",
                )
                for i in range(sensor_config.num_subsweeps)
            ]

            phase_curves.append((group_idx, sensor_id, subsweep_phase_curves))

            for phase_curve in subsweep_phase_curves:
                phase_plot.addItem(phase_curve)

            plot_widget.nextRow()

            subsweeps_distances_m = self.subsweeps_distances_m[group_idx][sensor_id]

            metadata = metadatas[group_idx][sensor_id]
            vels, vel_res = algo.get_approx_fft_vels(metadata, sensor_config)
            images = []
            for subsweep_index, subsweep_distances_m in enumerate(subsweeps_distances_m):
                step_length = sensor_config.subsweeps[subsweep_index].step_length
                plot = plot_widget.addPlot()
                plot.setMenuEnabled(False)
                plot.setLabel("bottom", "Distance (m)")
                plot.setLabel("left", "Velocity (m/s)")

                if sensor_config.num_subsweeps > 1:
                    plot.setTitle(f"Subsweep {subsweep_index + 1}")

                transform = QTransform()
                transform.translate(subsweep_distances_m[0], vels[0] - 0.5 * vel_res)
                transform.scale(metadata.base_step_length_m * step_length, vel_res)

                image = pg.ImageItem(autoDownsample=True)
                image.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
                image.setTransform(transform)

                plot.addItem(image)
                images.append(image)

            dvm_images.append((group_idx, sensor_id, images))

        self.phase_curves = core_utils.create_extended_structure(phase_curves)
        self.ft_images = core_utils.create_extended_structure(dvm_images)


class PluginSpec(
    ProcessorPluginSpec[
        List[Dict[int, a121.Result]],
        ProcessorConfig,
        ProcessorResult,
        List[Dict[int, a121.Metadata]],
    ]
):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


SPARSE_IQ_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="sparse_iq",
    title="Sparse IQ",
    description="Basic usage of the sparse IQ service.",
    family=PluginFamily.SERVICE,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2026
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et

from ._example_app import ExampleAppResult
from ._processor import ProcessorExtraResult, ProcessorResult


class VibrationPlot:
    """Vibration measurement plot handler"""

    def __init__(self) -> None:
        self._meas_dist_m: float = 0.0

    def setup_plot(self, widget: pg.GraphicsLayoutWidget, meas_dist_m: float) -> None:
        self._meas_dist_m = meas_dist_m
        pen_blue = et.utils.pg_pen_cycler(0)
        pen_orange = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(0)
        brush_dot = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw_blue = dict(pen=pen_blue, **symbol_kw)
        feat_kw_orange = dict(pen=pen_orange)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_dot, symbolPen="k")

        # presence plot
        self.object_detection_plot = pg.PlotItem()
        self.object_detection_plot.setMenuEnabled(False)
        self.object_detection_plot.showGrid(x=False, y=True)
        self.object_detection_plot.setLabel("left", "Max amplitude")
        self.object_detection_plot.setLabel("bottom", "Distance (m)")
        self.object_detection_plot.setXRange(meas_dist_m - 0.001, meas_dist_m + 0.001)
        self.presence_curve = self.object_detection_plot.plot(
            **dict(pen=pen_blue, **symbol_dot_kw)
        )

        self.presence_threshold = pg.InfiniteLine(pen=pen_blue, angle=0)
        self.object_detection_plot.addItem(self.presence_threshold)
        self.presence_threshold.show()

        self.smooth_max_presence = et.utils.SmoothMax(tau_decay=10.0)

        # sweep and threshold plot
        self.time_series_plot = pg.PlotItem()
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.setLabel("left", "Displacement (<font>&mu;</font>m)")
        self.time_series_plot.setLabel("bottom", "History")
        self.time_series_curve = self.time_series_plot.plot(**feat_kw_blue)

        self.time_series_plot.setYRange(-1000, 1000)
        self.time_series_plot.setXRange(0, 1024)

        self.text_item_time_series = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item_time_series.hide()
        self.time_series_plot.addItem(self.text_item_time_series)

        sublayout = widget.addLayout(row=0, col=0)
        sublayout.layout.setColumnStretchFactor(1, 5)
        sublayout.addItem(self.object_detection_plot, row=0, col=0)
        sublayout.addItem(self.time_series_plot, row=0, col=1)

        self.smooth_lim_time_series = et.utils.SmoothLimits(tau_decay=0.5, tau_grow=0.1)

        self.fft_plot = widget.addPlot(col=0, row=1)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.setLabel("left", "Displacement (<font>&mu;</font>m)")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.setLogMode(False, True)
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = [
            self.fft_plot.plot(**feat_kw_blue),
            self.fft_plot.plot(**feat_kw_orange),
            self.fft_plot.plot(**dict(pen=pen_blue, **symbol_dot_kw)),
        ]

        self.text_item_fft = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item_fft.hide()
        self.fft_plot.addItem(self.text_item_fft)

        self.smooth_max_fft = et.utils.SmoothMax()

    def update_plot(
        self,
        result: ExampleAppResult | ProcessorResult,
        extra_result: ProcessorExtraResult,
        meas_dist_m: float | None = None,
        show_time_series_std: bool = False,
    ) -> None:
        if meas_dist_m is None:
            meas_dist_m = self._meas_dist_m
        # Extra result
        time_series = extra_result.zm_time_series
        lp_displacements_threshold = extra_result.lp_displacements_threshold
        amplitude_threshold = extra_result.amplitude_threshold

        # Processor result
        lp_displacements = result.lp_displacements
        lp_displacements_freqs = result.lp_displacements_freqs
        max_amplitude = result.max_sweep_amplitude
        max_displacement = result.max_displacement
        max_displacement_freq = result.max_displacement_freq
        time_series_std = result.time_series_std

        # Plot object presence metric
        self.presence_curve.setData([meas_dist_m], [max_amplitude])
        self.presence_threshold.setValue(amplitude_threshold)
        lim = self.smooth_max_presence.update(max_amplitude)
        self.object_detection_plot.setYRange(0, max(1000.0, lim))

        # Plot time series
        if time_series is not None and amplitude_threshold < max_amplitude:
            assert time_series_std is not None
            lim = self.smooth_lim_time_series.update(time_series)
            self.time_series_plot.setYRange(lim[0], lim[1])
            self.time_series_plot.setXRange(0, time_series.shape[0])

            self.text_item_time_series.setPos(time_series.size / 2, lim[1] * 0.95)
            time_series_std_str = "{:.0f}".format(time_series_std)
            html_format = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:15pt;">'
                "{}</span></div>".format("STD : " + time_series_std_str + "<font>&mu;</font>m")
            )
            self.text_item_time_series.setHtml(html_format)
            if show_time_series_std:
                self.text_item_time_series.show()
            else:
                self.text_item_time_series.hide()
            self.time_series_curve.setData(time_series)

        # Plot spectrum
        if lp_displacements is not None:
            assert time_series is not None
            assert lp_displacements is not None

            self.fft_curve[0].setData(lp_displacements_freqs, lp_displacements)
            self.fft_curve[1].setData(lp_displacements_freqs, lp_displacements_threshold)
            lim = self.smooth_max_fft.update(np.max(lp_displacements))
            self.fft_plot.setYRange(-1, np.log10(lim))

            if max_displacement_freq is not None and max_displacement is not None:
                self.fft_curve[2].setData([max_displacement_freq], [max_displacement])

                # Place text box centered at the top of the plotting window
                self.text_item_fft.setPos(max(lp_displacements_freqs) / 2, np.log10(lim) * 0.95)
                html_format = (
                    '<div style="text-align: center">'
                    '<span style="color: #FFFFFF;font-size:15pt;">'
                    "{}</span></div>".format(
                        "Frequency: "
                        + str(int(max_displacement_freq))
                        + " Hz"
                        + "<br>"
                        + "Displacement: "
                        + str(int(max_displacement))
                        + " <font>&mu;</font>m"
                    )
                )
                self.text_item_fft.setHtml(html_format)
                self.text_item_fft.show()
            else:
                self.fft_curve[2].setData([], [])
                self.text_item_fft.hide()



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from ._configs import (
    get_high_frequency_config,
    get_low_frequency_config,
)
from ._example_app import (
    ExampleApp,
    ExampleAppConfig,
    ExampleAppResult,
    _load_algo_data,
)
from ._processor import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorExtraResult,
    ProcessorResult,
    ReportedDisplacement,
)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import copy
from typing import List, Optional

import attrs
import numpy as np
import numpy.typing as npt
from scipy.signal import filtfilt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    PERCEIVED_WAVELENGTH,
    AlgoProcessorConfigBase,
    PeakSortingMethod,
    ProcessorBase,
    _convert_amplitude_to_strength,
    get_distance_filter_coeffs,
    get_distance_filter_edge_margin,
    get_distance_offset,
    get_temperature_adjustment_factors,
)


TEMPERATURE_FILTER_CONSTANT = 0.95

# The same object can be seen at multiple subsweeps, objects close can be mereged.
MERGE_DISTANCE_M = 0.05
MERGE_SPEED_MPS = 0.05


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    num_std_treshold: float = attrs.field(default=5)
    num_mean_treshold: float = attrs.field(default=2)
    peak_sorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.STRONGEST,
        converter=PeakSortingMethod,
    )
    max_robot_speed: float = attrs.field(default=0.5)
    dead_reckoning_duration_s: float = attrs.field(default=0.5)
    sensitivity: float = attrs.field(default=0.5)

    def _collect_validation_results(
        self, config: Optional[a121.SessionConfig]
    ) -> list[a121.ValidationResult]:
        return []


@attrs.frozen(kw_only=True)
class Target:
    distance: float = attrs.field(default=None)
    velocity: float = attrs.field(default=None)
    strength: float = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class SubsweepProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    fft_map: npt.NDArray[np.float64] = attrs.field(default=None)
    fft_map_threshold: npt.NDArray[np.float64] = attrs.field(default=None)
    r: npt.NDArray[np.float64] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class SubsweepProcessorResult:
    targets: list[Target] = attrs.field(factory=list)
    extra_result: SubsweepProcessorExtraResult = attrs.field(factory=None)


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    dv: float = attrs.field(default=None)  # velocity difference between fft-bins


@attrs.frozen(kw_only=True)
class ProcessorResult:
    targets: list[Target] = attrs.field(factory=list)
    time: float = attrs.field(default=None)
    extra_result: ProcessorExtraResult = attrs.field(factory=ProcessorExtraResult)
    subsweeps_extra_results: List[SubsweepProcessorExtraResult] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class SubsweepProcessorContext:
    sub_sweep_idx: int = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class ProcessorContext:
    update_rate: float = attrs.field()
    mean_sweeps: list[npt.NDArray[np.float64]] = attrs.field(factory=list)
    std_sweeps: list[npt.NDArray[np.float64]] = attrs.field(factory=list)
    reference_temperature: Optional[float] = attrs.field(default=None)
    loopback_peak_location_m: float = attrs.field(default=0.0)


class SubsweepProcessor:
    """Obstacle processor

    :param sensor_config: Sensor configuration
    :param metadata: Metadata yielded by the sensor config
    :param processor_config: Processor configuration
    """

    LOOPBACK_START_IDX = -48

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        proc_context: ProcessorContext,
        ssproc_context: Optional[SubsweepProcessorContext] = None,
    ) -> None:
        self.proc_context = proc_context
        if ssproc_context is None:
            self.ssproc_context = SubsweepProcessorContext()
        else:
            self.ssproc_context = ssproc_context

        self.sensor_config = sensor_config
        self.processor_config = processor_config

        # Threshold setting
        self.num_std_threshold = processor_config.num_std_treshold
        self.num_mean_threshold = processor_config.num_mean_treshold

        # Calbration result
        self.true_zero_dist_idx = 0

        self.offset_m = get_distance_offset(
            self.proc_context.loopback_peak_location_m,
            sensor_config.profile,
            self.proc_context.reference_temperature,
        )

        if self.sensor_config.sweep_rate is None:
            self.sensor_config.sweep_rate = 100

        self.num_points = sensor_config.num_points
        self.dv = (
            PERCEIVED_WAVELENGTH
            / self.sensor_config.sweeps_per_frame
            * self.sensor_config.sweep_rate
        )
        (self.b, self.a) = get_distance_filter_coeffs(
            sensor_config.profile, sensor_config.step_length
        )
        self.fwhm_points = ENVELOPE_FWHM_M[sensor_config.profile] / (
            APPROX_BASE_STEP_LENGTH_M * sensor_config.step_length
        )

        self.filt_margin = get_distance_filter_edge_margin(
            sensor_config.profile,
            sensor_config.step_length,
        )

        # Extra
        self.r = APPROX_BASE_STEP_LENGTH_M * (
            sensor_config.start_point
            + self.filt_margin * sensor_config.step_length
            + sensor_config.step_length * np.arange(self.num_points - 2 * self.filt_margin)
        )

    def process(
        self, subframe: npt.NDArray[np.complex128], temperature: float
    ) -> SubsweepProcessorResult:
        assert self.proc_context.reference_temperature is not None

        # Depth filtering of each sweep
        filtered_subframe = self.apply_depth_filter(subframe)

        # Range downsampling and fft in the sweep dimension
        fftframe = np.fft.fft(filtered_subframe, axis=0)
        abs_fftframe = np.abs(fftframe)
        abs_fftframe_extra = np.copy(abs_fftframe)  # Copy for plotting

        sig_factor, noise_factor = get_temperature_adjustment_factors(
            reference_temperature=self.proc_context.reference_temperature,
            current_temperature=temperature,
            profile=self.sensor_config.profile,
        )

        sig_factor = (
            1 / sig_factor
        )  # After get_temperature_adjustment_factors updated, this inverted.

        fft_map_threshold = np.tile(
            noise_factor
            * self.num_std_threshold
            * self.proc_context.std_sweeps[self.ssproc_context.sub_sweep_idx]
            * np.sqrt(self.sensor_config.sweeps_per_frame),
            (self.sensor_config.sweeps_per_frame, 1),
        )
        fft_map_threshold[0, :] = (
            fft_map_threshold[0, :]
            + sig_factor
            * self.num_mean_threshold
            * np.abs(self.proc_context.mean_sweeps[self.ssproc_context.sub_sweep_idx])
            * self.sensor_config.sweeps_per_frame
        )

        targets = []

        bg_noise_stds = self.proc_context.std_sweeps[self.ssproc_context.sub_sweep_idx]

        diff = abs_fftframe - fft_map_threshold
        spf = self.sensor_config.sweeps_per_frame
        while np.any(diff > 0):
            idx_max = np.unravel_index(np.argmax(diff), diff.shape)
            i_dist = get_interpolated_range_peak_index(
                diff[idx_max[0], :]
            )  # A non-flat threshold can move a peak slightly
            i_speed = get_interpolated_fft_peak_index(fftframe[:, idx_max[1]], int(idx_max[0]))

            distance = (
                APPROX_BASE_STEP_LENGTH_M
                * (
                    self.sensor_config.start_point
                    + self.filt_margin * self.sensor_config.step_length
                    + self.sensor_config.step_length * i_dist
                )
                - self.offset_m
            )

            v = ((i_speed + spf / 2) % spf - spf / 2) * self.dv

            if 0 < idx_max[1] < (diff.shape[1] - 1):  # Disregard peaks at the limit of the range
                strength = _convert_amplitude_to_strength(
                    self.sensor_config.subsweeps[0],
                    abs_fftframe[idx_max[0], idx_max[1]],
                    distance,
                    bg_noise_stds[idx_max[1]],
                )
                targets.append(Target(distance=distance, velocity=v, strength=strength))

            abs_fftframe = subtract_reflector_from_fftmap(
                abs_fftframe, int(idx_max[1]), int(idx_max[0]), int(self.fwhm_points)
            )
            diff = abs_fftframe - fft_map_threshold

        er = SubsweepProcessorExtraResult(
            fft_map=abs_fftframe_extra, fft_map_threshold=fft_map_threshold, r=self.r
        )

        return SubsweepProcessorResult(targets=targets, extra_result=er)

    def apply_depth_filter(self, frame: npt.NDArray[np.complex128]) -> npt.NDArray[np.complex128]:
        # Written as a separate function to be callable during detector calibration

        filtered_array = np.array(filtfilt(self.b, self.a, frame), dtype=complex)
        cropped_array = filtered_array[:, self.filt_margin : -self.filt_margin]
        return cropped_array

    def update_config(self, config: ProcessorConfig) -> None:
        pass


class Processor(ProcessorBase[ProcessorResult]):
    """Obstacle processor

    :param sensor_config: Sensor configuration
    :param processor_config: Processor configuration
    :param context: Processor context
    """

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        processor_config: ProcessorConfig,
        context: ProcessorContext,
    ) -> None:
        self.sensor_config = sensor_config
        self.processor_config = processor_config

        self.num_subsweeps = sensor_config.num_subsweeps

        self.subsweep_processors: List[SubsweepProcessor] = []

        for i in range(self.num_subsweeps):
            subsweep_sensor_config_dict = sensor_config.to_dict()
            subsweep_sensor_config_dict["subsweeps"] = [sensor_config.subsweeps[i]]
            subsweep_sensor_config = a121.SensorConfig(**subsweep_sensor_config_dict)

            sspc = SubsweepProcessorContext(sub_sweep_idx=i)

            self.subsweep_processors.append(
                SubsweepProcessor(
                    sensor_config=subsweep_sensor_config,
                    processor_config=processor_config,
                    proc_context=context,
                    ssproc_context=sspc,
                )
            )

        # Calbration result
        self.true_zero_dist_idx = 0
        self.calbration_temperature = None

        self.filtered_sensor_temperature: Optional[float] = None
        self.temperature_filter_factor: float = 1 - np.exp(-TEMPERATURE_FILTER_CONSTANT)

        if self.sensor_config.sweep_rate is None:
            msg = "The obstacle detector needs the sweep_rate set."
            raise ValueError(msg)

        self.dv = (
            PERCEIVED_WAVELENGTH
            / 2
            / self.sensor_config.sweeps_per_frame
            * self.sensor_config.sweep_rate
        )

        self.kalman_filters: list[_KalmanFilter] = []
        self.update_rate = context.update_rate
        self.num_dead_reckoning_frames = int(
            processor_config.dead_reckoning_duration_s * self.update_rate
        )
        self.process_noise_gain_sensitivity = processor_config.sensitivity
        self.min_num_updates_valid_estimate = int(
            2 + (1 - self.process_noise_gain_sensitivity) * 20
        )

        # Largest anticipated distance change of target between frames. 4 to add margin.
        self.max_meas_state_diff_m = processor_config.max_robot_speed / self.update_rate * 4
        self.max_robot_speed = processor_config.max_robot_speed

    def process(self, result: a121.Result) -> ProcessorResult:
        # Filter temperature
        if self.filtered_sensor_temperature is None:
            self.filtered_sensor_temperature = float(result.temperature)
        else:
            self.filtered_sensor_temperature *= self.temperature_filter_factor
            self.filtered_sensor_temperature += (
                1 - self.temperature_filter_factor
            ) * result.temperature

        subsweep_results = [
            proc.process(subframe, temperature=self.filtered_sensor_temperature)
            for subframe, proc in zip(result.subframes, self.subsweep_processors)
        ]

        merged_targets = self._merge_subsweep_targets(subsweep_results)
        sorted_targets = self._sort_targets(
            merged_targets, self.processor_config.peak_sorting_method
        )

        # Update kalman filters.
        target_distances = [target.distance for target in sorted_targets]
        target_velocities = [target.velocity for target in sorted_targets]
        self.kalman_filters = self._update_kalman_filters(
            self.kalman_filters, target_distances, target_velocities
        )

        filtered_targets = []
        for target, kf in zip(sorted_targets, self.kalman_filters):
            distance = kf.get_distance()
            velocity = kf.get_velocity()
            new_target = Target(distance=distance, velocity=velocity, strength=target.strength)
            filtered_targets.append(new_target)

        subweeps_extra_results = [res.extra_result for res in subsweep_results]

        er = ProcessorExtraResult(dv=self.dv)

        return ProcessorResult(
            targets=filtered_targets,
            time=result.tick_time,
            extra_result=er,
            subsweeps_extra_results=subweeps_extra_results,
        )

    def _merge_subsweep_targets(
        self, subsweep_results: List[SubsweepProcessorResult]
    ) -> List[Target]:
        # The same object can be seen at multiple subsweeps, objects close can be mereged.

        all_targets = [target for sr in subsweep_results for target in sr.targets]

        while True:
            closest_dist = 2.0  # Initialized to larger than one
            for i in range(len(all_targets)):
                for j in range(i):
                    d = (
                        (all_targets[i].velocity - all_targets[j].velocity) / MERGE_SPEED_MPS
                    ) ** 2
                    d += (
                        (all_targets[i].distance - all_targets[j].distance) / MERGE_DISTANCE_M
                    ) ** 2

                    if d < closest_dist:
                        closest_dist = d
                        ij = (i, j)

            if closest_dist < 1.0:
                t1 = all_targets[ij[0]]
                t2 = all_targets[ij[1]]
                all_targets.append(
                    Target(
                        distance=(t1.distance + t2.distance) / 2,
                        velocity=(t1.velocity + t2.velocity) / 2,
                        strength=(t1.strength + t2.strength) / 2,
                    )
                )

                all_targets.remove(t1)
                all_targets.remove(t2)

            else:
                break

        return all_targets

    def apply_depth_filter(self, result: a121.Result) -> list[npt.NDArray[np.complex128]]:
        return [
            ssp.apply_depth_filter(r) for r, ssp in zip(result.subframes, self.subsweep_processors)
        ]

    def update_config(self, config: ProcessorConfig) -> None:
        pass

    @staticmethod
    def _sort_targets(
        targets: List[Target],
        method: PeakSortingMethod,
    ) -> List[Target]:
        if method == PeakSortingMethod.CLOSEST:
            quantity_to_sort = np.array([target.distance for target in targets])
        elif method == PeakSortingMethod.STRONGEST:
            quantity_to_sort = np.array([-target.strength for target in targets])
        else:
            msg = "Unknown peak sorting method"
            raise ValueError(msg)

        return [targets[i] for i in quantity_to_sort.argsort()]

    def get_closest_target_index(
        self,
        distances: npt.NDArray[np.float64],
        velocities: npt.NDArray[np.float64],
        kf_distance: float,
        kf_velocity: float,
    ) -> int:
        cost = np.sqrt(
            np.square((velocities - kf_velocity) / self.max_robot_speed)
            + np.square((distances - kf_distance) / self.max_meas_state_diff_m)
        )
        index = int(np.argmin(cost))

        return index

    def _update_kalman_filters(
        self,
        kfs: List[_KalmanFilter],
        distances: List[float],
        velocities: List[float],
    ) -> List[_KalmanFilter]:
        """Update Kalman filters for a sensor, using new distance and velocity estimates.
        Identifying the target, based on distance and velocity, closest to the current state of the filter.
        If the target is sufficiently close to the current state, it is used to update the filter.
        Once a target has been used to update a filter, it can't be used by other filters.
        If no estimated target matches the filter, dead reckoning is used. Each time dead
        reckoning is performed, a counter is incremented. If the counter exceeds a certain value,
        the filter is deleted.
        A filter must have a minimum number of updates before it is regarded as initiated.
        """
        distances = copy.deepcopy(list(distances))
        velocities = copy.deepcopy(list(velocities))
        kfs = copy.deepcopy(kfs)
        for kf in kfs:
            # Find the indexes of the closest distances.
            state_vs_distance_diff = np.abs(np.array(kf.get_distance()) - np.array(distances))
            idxs_close_to_estimate = np.where(state_vs_distance_diff < self.max_meas_state_diff_m)[
                0
            ]
            kf.predict()
            if len(distances) == 0 or len(idxs_close_to_estimate) == 0:
                # No estimates found.
                kf.dead_reckoning_count += 1
                # Remove the filter if not initialized or number of dead reckoning steps is to
                # high.
                if self.num_dead_reckoning_frames < kf.dead_reckoning_count or not kf.has_init:
                    kfs.remove(kf)
            else:
                # Estimates found.
                kf.dead_reckoning_count = 0
                # Find the target closest to the current estimates.
                distances_array = np.asarray(distances)
                velocities_array = np.asarray(velocities)
                idx_closest = self.get_closest_target_index(
                    distances_array[idxs_close_to_estimate],
                    velocities_array[idxs_close_to_estimate],
                    kf.get_distance(),
                    kf.get_velocity(),
                )
                idx_in_r_array = idxs_close_to_estimate[idx_closest]
                # Update KF
                kf.update(np.matrix([[distances[idx_in_r_array]], [velocities[idx_in_r_array]]]))
                # Remove distance and velocity as they have now been used to update a filter.
                distances.pop(idx_in_r_array)
                velocities.pop(idx_in_r_array)
                # Check if the minimum number of updates have been reached. If so, the filter has
                # been initialized.
                if self.min_num_updates_valid_estimate <= kf.num_updates:
                    kf.has_init = True
        for distance, velocity in zip(distances, velocities):
            kfs.append(
                _KalmanFilter(
                    1 / self.update_rate,
                    self.process_noise_gain_sensitivity,
                    distance,
                    velocity,
                    self.min_num_updates_valid_estimate,
                )
            )
        return kfs


def apply_max_depth_filter(
    abs_sweep: npt.NDArray[np.float64], config: a121.SubsweepConfig
) -> npt.NDArray[np.float64]:
    """Filter sweep so that every point is max within +/- one fwhm"""

    filtered_sweep = np.zeros_like(abs_sweep)

    half_max_length_idx = int(
        ENVELOPE_FWHM_M[config.profile] / (APPROX_BASE_STEP_LENGTH_M * config.step_length)
    )

    for i, _ in enumerate(abs_sweep):
        filtered_sweep[i] = np.max(
            abs_sweep[
                max(0, i - half_max_length_idx) : min(i + half_max_length_idx, len(abs_sweep))
            ]
        )

    return filtered_sweep


def get_interpolated_range_peak_index(env: npt.NDArray[np.float64]) -> float:
    """Simple quadratic peak interpolation assuming equidistant points"""

    SMALL_NUMBER = 1e-10  # To avoid divide by zero for a flat peak.

    i = int(np.argmax(env))

    if i == 0 or i == len(env) - 1:
        return i

    return float(
        i
        + (env[i + 1] - env[i - 1]) / (4 * env[i] - 2 * env[i - 1] - 2 * env[i + 1] + SMALL_NUMBER)
    )


def get_interpolated_fft_peak_index(Y: npt.NDArray[np.complex128], idx: int) -> float:
    """FFT peak interpolation according to Quinn's method"""

    Y2 = np.abs(Y[idx]) ** 2
    idxp1 = (idx + 1) % len(Y)
    ap = (Y[idxp1].real * Y[idx].real + Y[idxp1].imag * Y[idx].imag) / Y2
    dp = -ap / (1.0 - ap)
    am = (Y[idx - 1].real * Y[idx].real + Y[idx - 1].imag * Y[idx].imag) / Y2
    dm = am / (1.0 - am)

    return float(idx + (dp if ((dp > 0) & (dm > 0)) else dm))


def subtract_reflector_from_fftmap(
    fftmap: npt.NDArray[np.float64], r_idx: int, f_idx: int, fwhm: float
) -> npt.NDArray[np.float64]:
    """Subtract signal from single reflector in fft map"""

    MARGIN_FACTOR = 2

    Nf, Nr = fftmap.shape
    map_range = np.clip(
        1 - np.abs(np.arange(Nr) - r_idx) / (MARGIN_FACTOR * fwhm), 0, np.inf
    )  # Triangular envelope

    fs = np.arange(Nf)
    map_freq = np.sum([1 / (np.abs(f_idx + d * Nf - fs) + 1) for d in [-1, 0, 1]])

    peak = fftmap[f_idx, r_idx]

    return np.array(
        np.clip(
            fftmap - MARGIN_FACTOR * peak * np.outer(map_freq / np.max(map_freq), map_range),
            0,
            np.inf,
        )
    )


class _KalmanFilter:
    # Acceleration noise std (m/s^2).
    _PROCESS_NOISE_STD = 0.01
    # Distance estimated noise std (m).
    _MEASUREMENT_NOISE_STD = 0.005

    def __init__(
        self,
        dt: float,
        process_noise_gain_sensitivity: float,
        init_position: float,
        init_velocity: float,
        min_num_updates_valid_estimate: int,
    ) -> None:
        self.A: npt.NDArray[np.float64] = np.matrix([[1.0, dt], [0.0, 1.0]])
        self.H: npt.NDArray[np.float64] = np.matrix([[1, 0], [0, 1]])
        process_noise_gain = self._sensitivity_to_gain(process_noise_gain_sensitivity)
        # Random acceleration process noise.
        self.Q = (
            np.matrix([[(dt**4) / 4, (dt**3) / 2], [(dt**3) / 2, dt**2]])
            * (self._PROCESS_NOISE_STD) ** 2
            * process_noise_gain
        )
        self.R = self._MEASUREMENT_NOISE_STD**2
        self.P = np.eye(self.A.shape[1])
        self.x: npt.NDArray[np.float64] = np.matrix([[init_position], [init_velocity]])
        self.min_num_updates_valid_estimate = min_num_updates_valid_estimate
        self.dead_reckoning_count = 0
        self.num_updates = 0
        self.has_init = False

    def predict(self) -> None:
        self.x = np.dot(self.A, self.x)
        self.P = np.dot(np.dot(self.A, self.P), self.A.T) + self.Q

    def update(self, z: npt.NDArray[np.float64]) -> None:
        S = np.dot(self.H, np.dot(self.P, self.H.T)) + self.R
        K = np.dot(np.dot(self.P, self.H.T), np.linalg.inv(S))
        self.x = self.x + np.dot(K, (z - np.dot(self.H, self.x)))
        I = np.eye(self.H.shape[1])
        self.P = (I - (K * self.H)) * self.P
        self.num_updates += 1
        if self.min_num_updates_valid_estimate < self.num_updates:
            self.has_init = True

    def get_distance(self) -> float:
        return float(self.x[0, 0])

    def get_velocity(self) -> float:
        return float(self.x[1, 0])

    @staticmethod
    def _sensitivity_to_gain(sensitivity: float) -> float:
        return 0.01 + sensitivity * 20.0
